package PerlCAD::Color;

use vars qw(@ACI %ACI);
use strict;
use Exporter ();
use integer;

use Data::DumpEx;

@PerlCAD::Color::EXPORT = qw(@ACI %ACI
                             aci_to_name name_to_aci aci_to_int int_to_aci name_to_int int_to_name
                             validator);
@PerlCAD::Color::ISA = qw(Exporter);

###----------------------------------------------------------------###

sub validator {
  my $parent = shift || die caller;
  my $logic  = "ByLayer|ByBlock";
  my $named  = "Red|Green|Blue|Black|White|Cyan|Magenta|Yellow";
  my $html   = "#[a-f0-9]{6}";
  my $dec    = "\\d+,\\d+,\\d+";
  my $aci    = "[1-9]\\d?|1\\d\\d|2[0-5]\\d|25[1-5]"; # better done with QIntValidator - but we do all with on regex
  my $r = Qt::RegExp("^$logic|$named|$html|$dec|$aci\$", 0);
  return Qt::RegExpValidator($r, $parent);
}

###----------------------------------------------------------------###

### goes from named color to int
sub name_to_int {
  my $name = lc(shift || '');
  return [0,0,0] if $name eq 'byblock';
  return [0,0,0] if $name eq 'bylayer';
  return $ACI[$ACI{$name}] if $ACI{$name};
  $name =~ /^\#?([a-f0-9]{2})([a-f0-9]{2})([a-f0-9]{2})$/i || die "Invalid color name $name";
  return [map{hex} ($1, $2, $3)];
}

sub int_to_name {
  my @RGB = @_;
  return sprintf("#%02X%02X%02X", @RGB);
}

sub aci_to_name {
  my $i = shift;
  die "Bad value ($i)" if ! defined($i) || $i < 0 || $i > 256;
  return $ACI[$i] if $i == 0 || $i == 256;
  return sprintf("#%02X%02X%02X", @{ $ACI[$i] });
}

sub aci_to_int {
  my $i = shift;
  die "Bad value ($i)" if ! defined($i) || $i < 0 || $i > 256;
  return [0,0,0] if $i == 0 || $i == 256;
  return $ACI[$i];
}

sub name_to_aci {
  my $name = lc(shift || '');
  return 0   if $name eq 'byblock';
  return 256 if $name eq 'bylayer';
  return $ACI{$name} if $ACI{$name};
  $name =~ /^\#?([a-f0-9]{2})([a-f0-9]{2})([a-f0-9]{2})$/i || die "Invalid color name $name";
  return int_to_aci(map{hex} ($1, $2, $3));
}

sub int_to_aci {
  my @RGB = @_;

  ### allow for black on white
  return 7 if ! $RGB[0] && ! $RGB[1] && ! $RGB[2];

  ### first pass - direct comparison
  OUTER: for (my $i = 1; $i <= $#ACI - 1; $i ++) {
    for (0 .. 2) {
      next OUTER if $ACI[$i]->[$_] != $RGB[$_];
    }
    return $i;
  }

  ### next pass - get the closest index
  my $min = 16_700_000;
  my $min_i;
  OUTER: for (my $i = $#ACI - 1; $i >= 1; $i --) {
    my $sum = 0;
    $sum += ($ACI[$i]->[$_] - $RGB[$_])**2 for 0 .. 2;
    next if $sum > $min;
    $min = $sum;
    $min_i = $i;
  }
  die "Couldn't find an i for (@RGB)" if ! $min_i;
  return $min_i;
  
}

###----------------------------------------------------------------###

### standard ACI names
%ACI = (
        red     => 1,
        yellow  => 2,
        green   => 3,
        cyan    => 4,
        blue    => 5,
        magenta => 6,
        black   => 7,
        white   => 7,
        );

### AutoCad Color Index
@ACI = (
'byblock',       #0
[255, 0  , 0  ],
[255, 255, 0  ],
[0  , 255, 0  ],
[0  , 255, 255],
[0  , 0  , 255], #5
[255, 0  , 255],
[255, 255, 255],
[65 , 65 , 65 ],
[128, 128, 128],
[255, 0  , 0  ], #10
[255, 170, 170],
[189, 0  , 0  ],
[189, 126, 126],
[129, 0  , 0  ],
[129, 86 , 86 ], #15
[104, 0  , 0  ],
[104, 69 , 69 ],
[79 , 0  , 0  ],
[79 , 53 , 53 ],
[255, 63 , 0  ], #20
[255, 191, 170],
[189, 46 , 0  ],
[189, 141, 126],
[129, 31 , 0  ],
[129, 96 , 86 ], #25
[104, 25 , 0  ],
[104, 78 , 69 ],
[79 , 19 , 0  ],
[79 , 59 , 53 ],
[255, 127, 0  ], #30
[255, 212, 170],
[189, 94 , 0  ],
[189, 157, 126],
[129, 64 , 0  ],
[129, 107, 86 ], #35
[104, 52 , 0  ],
[104, 86 , 69 ],
[79 , 39 , 0  ],
[79 , 66 , 53 ],
[255, 191, 0  ], #40
[255, 234, 170],
[189, 141, 0  ],
[189, 173, 126],
[129, 96 , 0  ],
[129, 118, 86 ], #45
[104, 78 , 0  ],
[104, 95 , 69 ],
[79 , 59 , 0  ],
[79 , 73 , 53 ],
[255, 255, 0  ], #50
[255, 255, 170],
[189, 189, 0  ],
[189, 189, 126],
[129, 129, 0  ],
[129, 129, 86 ], #55
[104, 104, 0  ],
[104, 104, 69 ],
[79 , 79 , 0  ],
[79 , 79 , 53 ],
[191, 255, 0  ], #60
[234, 255, 170],
[141, 189, 0  ],
[173, 189, 126],
[96 , 129, 0  ],
[118, 129, 86 ], #65
[78 , 104, 0  ],
[95 , 104, 69 ],
[59 , 79 , 0  ],
[73 , 79 , 53 ],
[127, 255, 0  ], #70
[212, 255, 170],
[94 , 189, 0  ],
[157, 189, 126],
[64 , 129, 0  ],
[107, 129, 86 ], #75
[52 , 104, 0  ],
[86 , 104, 69 ],
[39 , 79 , 0  ],
[66 , 79 , 53 ],
[63 , 255, 0  ], #80
[191, 255, 170],
[46 , 189, 0  ],
[141, 189, 126],
[31 , 129, 0  ],
[96 , 129, 86 ], #85
[25 , 104, 0  ],
[78 , 104, 69 ],
[19 , 79 , 0  ],
[59 , 79 , 53 ],
[0  , 255, 0  ], #90
[170, 255, 170],
[0  , 189, 0  ],
[126, 189, 126],
[0  , 129, 0  ],
[86 , 129, 86 ], #95
[0  , 104, 0  ],
[69 , 104, 69 ],
[0  , 79 , 0  ],
[53 , 79 , 53 ],
[0  , 255, 63 ], #100
[170, 255, 191],
[0  , 189, 46 ],
[126, 189, 141],
[0  , 129, 31 ],
[86 , 129, 96 ], #105
[0  , 104, 25 ],
[69 , 104, 78 ],
[0  , 79 , 19 ],
[53 , 79 , 59 ],
[0  , 255, 127], #110
[170, 255, 212],
[0  , 189, 94 ],
[126, 189, 157],
[0  , 129, 64 ],
[86 , 129, 107], #115
[0  , 104, 52 ],
[69 , 104, 86 ],
[0  , 79 , 39 ],
[53 , 79 , 66 ],
[0  , 255, 191], #120
[170, 255, 234],
[0  , 189, 141],
[126, 189, 173],
[0  , 129, 96 ],
[86 , 129, 118], #125
[0  , 104, 78 ],
[69 , 104, 95 ],
[0  , 79 , 59 ],
[53 , 79 , 73 ],
[0  , 255, 255], #130
[170, 255, 255],
[0  , 189, 189],
[126, 189, 189],
[0  , 129, 129],
[86 , 129, 129], #135
[0  , 104, 104],
[69 , 104, 104],
[0  , 79 , 79 ],
[53 , 79 , 79 ],
[0  , 191, 255], #140
[170, 234, 255],
[0  , 141, 189],
[126, 173, 189],
[0  , 96 , 129],
[86 , 118, 129], #145
[0  , 78 , 104],
[69 , 95 , 104],
[0  , 59 , 79 ],
[53 , 73 , 79 ],
[0  , 127, 255], #150
[170, 212, 255],
[0  , 94 , 189],
[126, 157, 189],
[0  , 64 , 129],
[86 , 107, 129], #155
[0  , 52 , 104],
[69 , 86 , 104],
[0  , 39 , 79 ],
[53 , 66 , 79 ],
[0  , 63 , 255], #160
[170, 191, 255],
[0  , 46 , 189],
[126, 141, 189],
[0  , 31 , 129],
[86 , 96 , 129], #165
[0  , 25 , 104],
[69 , 78 , 104],
[0  , 19 , 79 ],
[53 , 59 , 79 ],
[0  , 0  , 255], #170
[170, 170, 255],
[0  , 0  , 189],
[126, 126, 189],
[0  , 0  , 129],
[86 , 86 , 129], #175
[0  , 0  , 104],
[69 , 69 , 104],
[0  , 0  , 79 ],
[53 , 53 , 79 ],
[63 , 0  , 255], #180
[191, 170, 255],
[46 , 0  , 189],
[141, 126, 189],
[31 , 0  , 129],
[96 , 86 , 129], #185
[25 , 0  , 104],
[78 , 69 , 104],
[19 , 0  , 79 ],
[59 , 53 , 79 ],
[127, 0  , 255], #190
[212, 170, 255],
[94 , 0  , 189],
[157, 126, 189],
[64 , 0  , 129],
[107, 86 , 129], #195
[52 , 0  , 104],
[86 , 69 , 104],
[39 , 0  , 79 ],
[66 , 53 , 79 ],
[191, 0  , 255], #200
[234, 170, 255],
[141, 0  , 189],
[173, 126, 189],
[96 , 0  , 129],
[118, 86 , 129], #205
[78 , 0  , 104],
[95 , 69 , 104],
[59 , 0  , 79 ],
[73 , 53 , 79 ],
[255, 0  , 255], #210
[255, 170, 255],
[189, 0  , 189],
[189, 126, 189],
[129, 0  , 129],
[129, 86 , 129], #215
[104, 0  , 104],
[104, 69 , 104],
[79 , 0  , 79 ],
[79 , 53 , 79 ],
[255, 0  , 191], #220
[255, 170, 234],
[189, 0  , 141],
[189, 126, 173],
[129, 0  , 96 ],
[129, 86 , 118], #225
[104, 0  , 78 ],
[104, 69 , 95 ],
[79 , 0  , 59 ],
[79 , 53 , 73 ],
[255, 0  , 127], #230
[255, 170, 212],
[189, 0  , 94 ],
[189, 126, 157],
[129, 0  , 64 ],
[129, 86 , 107], #235
[104, 0  , 52 ],
[104, 69 , 86 ],
[79 , 0  , 39 ],
[79 , 53 , 66 ],
[255, 0  , 63 ], #240
[255, 170, 191],
[189, 0  , 46 ],
[189, 126, 141],
[129, 0  , 31 ],
[129, 86 , 96 ], #245
[104, 0  , 25 ],
[104, 69 , 78 ],
[79 , 0  , 19 ],
[79 , 53 , 59 ],
[51 , 51 , 51 ], #250
[80 , 80 , 80 ],
[105, 105, 105],
[130, 130, 130],
[190, 190, 190],
[255, 255, 255], #255
'bylayer',       #256
  );

###----------------------------------------------------------------###

1;
