package PerlCAD::Command;

###----------------------------------------------------------------###
#  Copyright 2003 - Paul Seamons                                     #
#  Distributed under the GNU General Public License without warranty #
###----------------------------------------------------------------###

use vars qw($loaded
            %REQUIRED
            %COMMANDS
            %ALIASES
            %ACTIONS
            $VOID_N
            %IS_GUI
            %REGISTERED_SIG
            %REGISTERED_SLOT
            $QR_CHUNK_FROM_LINE
            $QR_NUMBER
            $no_recurse
            );
use strict;
use Qt;
use Qt::isa qw(Qt::Object);
use Qt::signals
  sig_display_line   => ['QString', 'int', 'int'],
  sig_send_prompt    => [],
  sig_relative_point => ['double', 'double', 'double'],
  sig_script_write_stdin => ['const QString &'],
  ;
use Qt::slots
  slot_process_line   => ['QString', 'int'],
  slot_escape_pressed => [],
  slot_script_exited  => [],
  slot_script_ready_read => [],
  ;
use Qt::attributes qw(
                      state
                      multiple_list

                      active_command
                      active_alias
                      active_context
                      active_option
                      default_option
                      requires_queue

                      history
                      last_point
                      last_distance
                      dev_no_counter
                      dedicated_device
                      view

                      active_script
                      );
use PerlCAD::Icon;
use PerlCAD::Dump qw(debug);
use Math::Trig ();

use constant {
  GENERAL => 0,
  INFO    => 1,
  WARNING => 2,
  ERROR   => 3,
};
use constant max_history => 100;

$IS_GUI{0}  = 1; # __PACKAGE__ is a GUI
$IS_GUI{-1} = 1; # recurred command is gui
$VOID_N     = 0;
$QR_CHUNK_FROM_LINE = qr/^\s*([^\s;]+)\s*($|;|\s)/;
$QR_NUMBER  = qr/[+-]?(?:\d+|\d*\.\d+)/;

###----------------------------------------------------------------###

sub NEW {
  shift->SUPER::NEW(@_[0,1]);
  set_last_point([0, 0, 0]);
  clear_history();
  state = {};
  multiple_list = [];
}

sub set_last_point {
  my $point  = shift;
  last_point = $point;
  send_relative_point($point);
}

sub send_relative_point {
  my $point = shift;
  $point->[2] ||= 0;
  emit sig_relative_point(@$point);
}

sub clear_history {
  history = {};
}

sub get_state { state }

sub set_state { state = shift }

###----------------------------------------------------------------###

sub load_commands_and_actions {
  return if $loaded;
  $loaded = 1;

  my $subdir = __PACKAGE__;
  $subdir =~ s|::|/|g;
  foreach my $dir (@INC) {
    opendir(_DH,"$dir/$subdir") || next;
    #print "Found command dir: $dir/$subdir\n";
    while (defined(my $file = readdir(_DH))) {
      next if $file !~ /^\w+.pm$/;
      next if $REQUIRED{$file};
      $REQUIRED{$file} = 1;
      my $file = "$dir/$subdir/$file";
      if (! eval {require $file}) {
        warn "Couldn't require $file: $@";
      }
    }
    closedir(_DH);
  }

}

sub load_aliases {
}

###----------------------------------------------------------------###

### register_command stores a method that will be called from the context
### of PerlCAD::Command for any given command
sub register_command {
  my $args = shift;
  my $command = uc($args->{command}) || die "Usage ({command => command, method => sub {}})";

  if ($COMMANDS{$command}) {
    warn "A command already exists for $command - using the first\n";
    return 0;
  } elsif ($command =~ /\s/) {
    warn "Commands may not contain whitespace ($command)";
    return 0;
  } elsif ($command =~ /^([_+-])/) {
    warn "Commands should not begin with a _, +, or - ($command)";
    return 0;
  }

  if ($args->{options}) {
    die "Options for command $command must be a hashref or a coderef"
      if ref($args->{options}) ne 'HASH' && ref($args->{options}) ne 'CODE';
  } elsif ($args->{method}) {
    die "Method for command $command must be a coderef"
      if ref($args->{method}) ne 'CODE';
  } else {
    die "Not sure how to use command $command (no options or method found)";
  }

  ### all OK - save the args
  $COMMANDS{$command} = $args;
  if ($args->{aliases}) {
    my $ref = ref($args->{aliases}) ? $args->{aliases} : [$args->{aliases}];
    $ALIASES{uc($_)} = $command foreach @$ref;
  }
  return 1;
}


### register_action is used to create a Qt::Action for the given action
### it takes the name of the action, a string for the method to call
### a possible tooltip and an optional whatsthis string
sub register_action {
  my $args = shift;

  ### set up a qt action for an item
  my $name = $args->{name} || die "Invalid or missing name";
  my $obj = $ACTIONS{$name} = Qt::Action(this, "${name}_action", $args->{is_toggle});
  $obj->setOn(1) if $args->{is_toggle} && $args->{toggled};

  ### get themed icons or fall back to internal defaults
  my $dirs  = this->parent->config->icon_dirs();
  my $img;
  foreach my $dir (@$dirs) {
    foreach my $ext (qw(png xpm)) {
      my $file = "$dir/$name.$ext";
      next if ! -e $file;
      $img = Qt::Pixmap();
      $img->load($file);
      last;
    }
  }
  $img ||= &PerlCAD::Icon::data($name);
  $obj->setIconSet(Qt::IconSet(Qt::Pixmap($img))) if $img;

  ### set the accel (key sequence)
  ### can be either a key sequence string, or an array ref of key codes
  my $accel = $args->{accel} || '';
  $obj->setAccel(Qt::KeySequence(ref($accel) ? @$accel : $accel)) if $accel;

  ### set all of the texts
  my $menutext = $args->{menutext};
  if (! $menutext) {
    $menutext = $name;
    $menutext =~ s/^[^_]+_//;
    $menutext =~ y/_/ /;
    $menutext =~ s/\b(\w)/\u$1/g;
  }
  $obj->setMenuText($menutext);

  my $tooltext = $args->{tooltext};
  if (! defined $tooltext) {
    $tooltext = $menutext;
    $tooltext =~ tr/&//d;
  }
  $obj->setText($tooltext);

  my $tooltip = $args->{tooltip} || '';
  $tooltip .= "($name not implemented)" if ! $args->{slot} || ! $args->{method};
  $obj->setToolTip($tooltip) if $tooltip;

  my $statustip = $args->{statustip} || $tooltip;
  $obj->setStatusTip($statustip) if $statustip;

  my $whatsthis = $args->{whatsthis};
  if ($whatsthis) {
    if ($whatsthis =~ /<img \s+ src= ['"] (\w+) ["']/ix) {
      my $pixmap = $obj->iconSet->pixmap;
      Qt::MimeSourceFactory::defaultFactory()->setPixmap($1, $pixmap) if $pixmap;
    }
    $obj->setWhatsThis($whatsthis);
  }

  ### connect the signal
  my $slot   = $args->{slot} || 'void_slot_'. $VOID_N ++;
  my $method = $args->{method};
  if ($args->{is_toggle}) {
    $method ||= sub { my $val = shift || 0; warn "Void toggle action called \"$name\" = $val\n"; };
    import Qt::slots $slot => ['bool'];
    this->connect($obj, SIGNAL('toggled(bool)'), SLOT("${slot}(bool)"));
  } else {
    $method ||= sub { warn "Void action called \"$name\"\n" };
    import Qt::slots $slot => [];
    this->connect($obj, SIGNAL('activated()'), SLOT("${slot}()"));
  }
  no strict 'refs';
  *{__PACKAGE__ . "::" . $slot} = $method;
  use strict 'refs';
}

###----------------------------------------------------------------###

sub get_action {
  my $action = shift;
  if (! exists $ACTIONS{$action}) {
    register_action({name => $action});
  }
  return $ACTIONS{$action};
}

###----------------------------------------------------------------###

sub register_device {
  my $device    = shift;
  my $sig_name  = shift;
  my $slot_name = shift;
  my $is_gui    = shift;

  dev_no_counter ||= 0;
  dev_no_counter ++;
  my $dev_no = dev_no_counter;
  $device->set_dev_no($dev_no);

  if ($sig_name) {
    this->connect($device, SIGNAL("${sig_name}(QString, int)"), SLOT('slot_process_line(QString,int)'));
    $REGISTERED_SIG{$dev_no} = $sig_name;
  }
  if ($slot_name) {
    $device->connect(this, SIGNAL("sig_display_line(QString,int,int)"), SLOT("${slot_name}(QString,int,int)"));
    $REGISTERED_SLOT{$dev_no} = $slot_name;
  }

  set_is_gui($dev_no, $is_gui);
}

sub unregister_device {
  my $device = shift;
  my $dev_no = $device->get_dev_no || return;

  if ($REGISTERED_SIG{$dev_no}) {
    my $sig_name = delete $REGISTERED_SIG{$dev_no};
    this->disconnect($device, SIGNAL("${sig_name}(QString, int)"));
  }
  if ($REGISTERED_SLOT{$dev_no}) {
    my $slot_name = delete $REGISTERED_SLOT{$dev_no};
    $device->disconnect(this, SIGNAL("sig_display_line(QString,int,int)"));
  }

  set_is_gui($dev_no, 0);
}

sub set_is_gui {
  my $dev_no = shift || die "Missing dev_no";
  if (shift) {
    $IS_GUI{$dev_no} = 1;
  } else {
    delete $IS_GUI{$dev_no};
  }
}

sub is_gui {
  my $dev_no = shift || return 1;
  return $IS_GUI{$dev_no};
}

sub set_dedicated {
  my $device = shift;
  my $dev_no = $device->get_dev_no || die "Missing dev_no";
  die "Cannot set multiple dedicated devices" if dedicated_device;
  dedicated_device = $dev_no;
}

sub unset_dedicated {
  dedicated_device = undef;
  send_prompt();
}

###----------------------------------------------------------------###

sub emit_general {
  my $msg    = shift;
  my $dev_no = shift || 0;
  return if ! defined($msg) || ! length($msg);
  $msg =~ s/^\s*/>> /mg if state->{is_transparent};
  emit sig_display_line($msg, $dev_no, GENERAL);
  emit sig_script_write_stdin($msg) if active_script;
  return 1;
}

sub emit_warning {
  my $msg    = shift || return;
  my $dev_no = shift || 0;
  $msg = "<$msg>";
  $msg =~ s/^\s*/>> /mg if state->{is_transparent};
  emit sig_display_line($msg, $dev_no, WARNING);
  return 1;
}

sub emit_info {
  my $msg    = shift || return;
  my $dev_no = shift || 0;
  $msg = "*$msg*";
  $msg =~ s/^\s*/>> /mg if state->{is_transparent};
  emit sig_display_line($msg, $dev_no, INFO);
  return 1;
}

sub send_prompt {
  emit sig_send_prompt();
}

###----------------------------------------------------------------###

sub slot_process_line {
  my $line   = shift;
  my $dev_no = shift || 0;
  my $state  = state ||= {};
  my $started_this_time = 0;

  ### allow for some commands which may call themselves (toggle actions)
  if ($state->{force_no_recurse}) {
    #dex "Not recursed $line";
    return;
  }

  ### don't need whitespace
  $line =~ s/\s+$//;
  $line =~ s/^\s+//;
  $state->{line_was_empty} = ! length($line);

  ### don't allow info from others during dedicated mode
  if (dedicated_device) {
    return if dedicated_device != $dev_no;
    emit_info($line, $dev_no);
  }

  ### if we have a running command - spit out the input 
  if ($state->{command} && $line !~ /^\'/) {
    if ($state->{is_multiple}) {
      $state->{is_multiple}    = undef;
      $state->{line_was_empty} = undef;
      $no_recurse = 0;
    } else {
      emit_general($line, $dev_no);
    }

  } else {

    ### verify transparent command status
    if ($state->{command}) {
      if ($line !~ /^\'/) {
        emit_info("Transparent command attempt without a quote");
        send_prompt();
        return;
      } elsif ($state->{is_transparent}) {
        emit_warning(sprintf("Transparent command \"%s\" already in progress ($line)",$state->{command}));
        send_prompt();
        return;
      }
    }

    ### if no command passed - recall one from history
    my $hist = history()->{'__commands'} ||= [];
    if ($state->{line_was_empty} && $hist && $#$hist != -1) {
      $state->{line_was_empty} = undef;
      $line = $hist->[0];
      $line = $1 if $line =~ m/^\s*([^;]+)\s*($|;|\s)/;
      $dev_no = -1; # allow recurred commands to show up
    }

    ### get the command
    if ($line !~ s/$QR_CHUNK_FROM_LINE//) {
      send_prompt();
      return;
    }
    my $command = uc($1);
    my $context = ($command =~ s/^([_\-\']+)//) ? $1 : '';
    my $alias   = '';
    if ($ALIASES{$command}) {
      $alias   = $command;
      $command = $ALIASES{$command};
    }

    ### save command history - even if it is a bad command
    if (! $state->{command}
        && $command !~ /^\d/) {
      my $save = lc($command);
      $save .= " $line" if $line =~ /\S/;
      $save = "-$command" if index($context,'-') != -1;
      if (! defined($hist->[0]) || $hist->[0] ne $save) {
        pop @$hist if $#$hist >= max_history - 1;
        unshift @$hist, $save;
      }
    }

    ### report the command to the user
    if (! $COMMANDS{$command}) {
      emit_warning("Unknown command \"$command\"");
      send_prompt();
      return;
    }

    ### verify context and transparency
    if ($state->{command}) {
      $context = "'$context" if index($context,"'") == -1 && $IS_GUI{$dev_no};
      if (index($context, "'") == -1) {
        emit_warning(sprintf("Command \"%s\" called with invalid context",$command));
        $line = '';
        check_requires($state,\$line);
        return;
      } elsif (! $COMMANDS{$command}->{allow_transparent}) {
        emit_warning(sprintf("Command \"%s\" is not a transparent command",$command));
        $line = '';
        check_requires($state,\$line);
        return;
      }
    } else {
      $context =~ s/\'//; # nicely unset transparent context if we aren't
    }

    ### set up new state
    my $previous_state = $state;
    $state = {
      command => $command,
      alias   => $alias,
      context => $context,
      line_was_empty => $state->{line_was_empty},
      use_dialog     => $COMMANDS{$command}->{has_dialog} && index($context,'-') == -1,
    };
    if (index($context,"'") != -1) {
      $state->{is_transparent} = 1;
      $state->{previous_state} = $previous_state;
    }
    state = $state;
    $started_this_time = 1;

    ### spit out the command
    my $str = "${context}${command}";
    $str .= " $line" if $line =~ /\S/;
    emit_general($str, $dev_no);

  }


  ### allow for toggle switches that could retrigger themselves
  return if $no_recurse;
  local $no_recurse = 1;


  ### by now we should have a command
  ### now check for required arguments to the command
  my $command = $state->{command};
  my $queue   = $state->{queue} ||= [];
  my $c_args  = $COMMANDS{$command};
  my $allow_multiple;

  ### run options
  if (! $state->{use_dialog} && $c_args->{options}) {
    my $option;
    my $options = $c_args->{options};
    $options = &$options() if ref($options) eq 'CODE';

    ### check for active or default option
    if ($state->{option}) {
      $option = $state->{option};

    ### if the line was empty - look for default (otherwise cancel the command)
    } elsif ($state->{line_was_empty}) {
      if ($state->{default}) {
        $option = $state->{default}->{option};
      } else {
        emit_info("Cancelled");
      }

    ### allow for choosing a point or distance to activate the default
    } elsif ($state->{default} && $state->{default}->{activate_on}) {
      if ($state->{default}->{activate_on} eq 'POINT') {
        my $point = eval { parse_point($line, 1) };
        if ($point) {
          $option = $state->{default}->{option};
          set_last_point($point);
          push @$queue, $point;
          $line =~ s/^\s*[^;]+\s*($|;)//;
          $state->{save} = join(",", @$point);
        }
      } elsif ($state->{default}->{activate_on} eq 'DISTANCE') {
        my $point = eval { parse_point($line, 1) };
        if ($point) {
          $option = $state->{default}->{option};
          set_last_point($point);
          push @$queue, $point;
          $line =~ s/^\s*[^;]+\s*($|;)//;
          $state->{save} = join(",", @$point);
        } else {
          my $dist = eval { parse_number($line, 1) };
          if (defined $dist) {
            $option = $state->{default}->{option};
            push @$queue, $dist;
            $line =~ s/^\s*[^;]+\s*($|;)//;
            $state->{save} = $dist;
          }
        }
      } else {
        die "Unknown activate_on option \"$state->{default}->{activate_on}\"";
      }
    }

    ### look for valid option by name or key
    if (! $option) {
      if ($line =~ s/^\s*([^\s;]+)\s*//) {
        my $test = $1;
        if ($options->{uc($test)}) {
          $option = uc($test);
        } else {
          my %KEYS = map {$options->{$_}->{key} => $_} grep {$options->{$_}->{key}} keys %$options;
          if ($KEYS{uc($test)}) {
            $option = $KEYS{uc($test)};
          } else {
            emit_warning("Invalid option \"$test\"");
          }
        }
      }
      if (! $option) {
        return show_usage($c_args);
      }
    }

    ### run the option if we have it
    if ($option) {
      if (! $state->{option}) {
        $state->{option} = $option;
        ### save history
        my $save = delete($state->{save}) || lc($option);
        my $ref = history()->{'__options'}->{$command} ||= [];
        if (! defined($ref->[0]) || $ref->[0] ne $save) {
          pop @$ref if $#$ref == 99;
          unshift @$ref, $save;
        }
      }
      my $coderef = $options->{$option}->{method} || die "Missing method for \"$command $option\"";
      $state->{requires} ||= $options->{$option}->{requires};
      if (! check_requires($state, \$line)) {
        return if ! $state->{line_was_empty};
        emit_info("Done");
      } else {
        $state->{force_no_recurse} = 1 if $options->{$option}->{no_recurse};
        eval { &$coderef($state, @$queue) };
        if ($@) {
          my $err = "$@";
          chomp $err;
          emit_info($err);
        } else {
          $allow_multiple = $options->{$option}->{multiple};
        }
      }
    }

  ### run single method (no options on this command)
  } elsif ($c_args->{method}) {
    my $coderef = $c_args->{method};
    $state->{requires} ||= $c_args->{requires};
    if (! $state->{use_dialog} && ! check_requires($state, \$line)) {
      return if ! $state->{line_was_empty};
      emit_info("Done");
    } else {
      $state->{force_no_recurse} = 1 if $c_args->{no_recurse};
      &$coderef($state, @$queue);
      $allow_multiple = $c_args->{multiple};
    }
  }


  ### restore from a transparent command
  if ($state->{is_transparent}) {
    state = $state->{previous_state};
    $line = '';
    check_requires(state, \$line) && send_prompt();
    return;

  ### allow for commands that repeat
  } elsif ($allow_multiple && ! $started_this_time) {
    push @{ multiple_list() }, {
      command => $state->{command},
      option  => $state->{option},
      queue   => $state->{queue},
    };
    state = {
      command  => $command,
      alias    => state->{alias},
      context  => state->{context},
      is_multiple => 1,
    };
    &$allow_multiple() if ref($allow_multiple) eq 'CODE';
    slot_process_line($line, $dev_no);

  ### otherwise - clear the command
  } else {
    state = {};
    multiple_list = [];
    send_prompt();
  }

}

### see if this option requires any additional input
sub check_requires {
  my $state    = shift;
  my $lineref  = shift;
  my $requires = $state->{requires} || return 1;
  my $queue    = $state->{queue} ||= [];
  $requires = &$requires() if ref($requires) eq 'CODE';
  die "Invalid requires - must be an arrayref" if ref($requires) ne 'ARRAY';

  ### iterate on the required items making sure we have a queued piece
  for (my $i = 0; $i <= $#$requires; $i ++) {
    my $hash = $requires->[$i];
    $hash = &$hash() if ref($hash) eq 'CODE';
    die "Invalid requires item - must be a hashref" if ref($hash) ne 'HASH';
    die "Missing type" if ! $hash->{type};

    my $type = uc($hash->{type});
    my $msg  = $hash->{msg} || '';
    my $found;
    my $default;

    ### allow for default to be placed in queue
    ### in theory - the item shouldn't succeed on first pass (such as point for distance)
    if (! defined($queue->[$i]) && $hash->{default_queue}) {
      my $item = $hash->{default_queue};
      $item = &$item() if ref($item) eq 'CODE';
      $queue->[$i] = $item;
    }

    ### one line of text
    if ($type eq 'TEXTLINE') {
      if (defined($queue->[$i]) && length($queue->[$i])) {
        $found = 1;
      } elsif ($$lineref =~ s/^\s*(.+)\s*$//) {
        $queue->[$i] = $1;
        save_requires_history($type, $1);
        $found = 1;
      }

    ### look for a single point
    } elsif ($type eq 'POINT') {
      if (defined $queue->[$i]) {
        my $point = $queue->[$i];
        die "Invalid point in queue (@$point)" if $#$point > 2 || $#$point < 1;
        $found = 1;
      } elsif ($state->{line_was_empty}) {
        if ($hash->{default_to_last}) {
          $state->{line_was_empty} = undef;
          $queue->[$i] = last_point;
          $found = 1;
        }
      } elsif ($$lineref =~ s/$QR_CHUNK_FROM_LINE//) {
        # could do .x, .y, .z, .xy, .xz, .yz, .xyz here
        my $parts = eval { parse_point($1) };
        if ($parts) {
          $queue->[$i] = $parts;
          set_last_point($parts);
          save_requires_history($type, join(",", @$parts));
          $found = 1;
        }
      }
      $default = join(",",@{ last_point() }) if ! $found && $hash->{default_to_last};

    ### allow entry of distance - done via length or two points
    } elsif ($type eq 'DISTANCE') {
      my $dist;
      if ($queue->[$i] && ! ref($queue->[$i])) {
        $found = 1;
      } elsif ($state->{line_was_empty}) {
        if ($hash->{default_to_last}) {
          $state->{line_was_empty} = undef;
          $dist = last_distance;
        } elsif ($hash->{default}) {
          $state->{line_was_empty} = undef;
          $dist = $hash->{default};
          $dist = &$dist() if ref($dist) eq 'CODE';
        }
      } else {
        if ($$lineref =~ s/$QR_CHUNK_FROM_LINE//) {
          my $chunk = $1;
          my $p1    = eval { parse_point($chunk, 1) };
          if ($p1) { # has point
            my $p2 = $queue->[$i];
            if ($p2 && ref($p2)) { # has two points
              $_->[2] ||= 0 foreach $p1, $p2;
              $dist = sqrt(  ($p1->[0] - $p2->[0])**2
                           + ($p1->[1] - $p2->[1])**2
                           + ($p1->[2] - $p2->[2])**2);
              # dex $dist;  # seems to report a little low
              $dist = sprintf("%.".this->parent->sig_fig."f", $dist);
              set_last_point($p1);
              save_requires_history('POINT', join(",", @$p1));
            } else { # save one point for next time
              $queue->[$i] = $p1;
            }
          } else {
            $dist = eval { parse_number($chunk) };
          }
        }
        if (! defined($dist) && ref($queue->[$i])) {
          $msg ||= "Please enter second point for distance";
        }
      }
      if (defined($dist)) {
        $queue->[$i]  = $dist;
        last_distance = $dist;
        save_requires_history($type, $dist);
        $found = 1;
      }
      $default ||= last_distance if ! $found && $hash->{default_to_last} && defined(last_distance);

    } elsif ($type eq 'SELECTION') {

    } elsif ($type eq 'UINT') {
      if (defined($queue->[$i])) {
        $found = 1;
      } elsif ($$lineref =~ s/$QR_CHUNK_FROM_LINE//) {
        my $int = $1;
        if ($int =~ /^(0|[1-9]\d*)$/) {
          $queue->[$i] = $int;
          save_requires_history($type, $int);
          $found = 1;
        }
      }

    } elsif ($type eq 'REAL') {
      if (defined($queue->[$i])) {
        $found = 1;
      } elsif ($$lineref =~ s/$QR_CHUNK_FROM_LINE//) {
        my $real = $1;
        if ($real =~ /^(\d+|\d*.\d+)$/) {
          $queue->[$i] = 1 * $real;
          save_requires_history($type, $real);
          $found = 1;
        }
      }

    } elsif ($type eq 'FILENAME') {
      if (defined($queue->[$i])) {
        $found = 1;
      } elsif ($$lineref =~ s/$QR_CHUNK_FROM_LINE//) {
        my $file = $1;
        if (! -e $file) {
          emit_warning("File does not exist \"$file\"");
        } else {
          $queue->[$i] = $file;
          save_requires_history($type, $file);
          $found = 1;
        }
      }

    } elsif ($type eq 'COLOR') {
      if (defined($queue->[$i])) {
        $found = 1;
      } elsif ($$lineref =~ s/$QR_CHUNK_FROM_LINE//) {
        my $name = $1;
        my $v = &PerlCAD::Color::validator(this);
        my $n = 0;
        if ($v->validate($name, $n) == Qt::Validator::Acceptable()) {
          $queue->[$i] = $name;
          save_requires_history($type, $name);
          $found = 1;
        } else {
          emit_warning("Invalid color \"$name\"");
        }
      }
      $msg ||= "Specify a color [ACI value/ByLayer/ByBlock/#hexval]";

    } else {
      die "Unknown required type $type";
    }

    ### show a message if nothing is found
    if (! $found) {
#      debug $hash;
      if ($hash->{optional}) {
        this->parent->doc->active_view()->set_draw_sample();
        return 1; # assumes that all options after first optional item - are optional
      } else {
        if (! $default && exists($hash->{default})) {
          $default = $hash->{default};
          $default = &$default() if ref($default) eq 'CODE';
        }
        $msg = &$msg() if ref($msg) eq 'CODE';
        $msg ||= "Please enter a" . " \u\L$type";
        $msg =~ s/\:?\s*//;
        $msg .= " <$default>" if defined($default) && length($default);
        $msg .= ": ";
        if (! $state->{line_was_empty}) {
          emit_general($msg);
          if ($hash->{draw}) {
            this->parent->doc->active_view()->set_draw_sample($hash->{draw}, $queue);
          } else {
            this->parent->doc->active_view()->set_draw_sample();
          }
        } else {
          this->parent->doc->active_view()->set_draw_sample();
        }
        return 0;
      }
    } else {
      $state->{highest_found} = $i;
    }
  }

  ### all options accounted for
  this->parent->doc->active_view()->set_draw_sample();
  return 1;
}

###----------------------------------------------------------------###

### get a point from the passed in line
sub parse_point {
  my $line = shift;
  my $die_on_error = shift;
  my $is_relative = $line =~ s/^\s*\@\s*//;
  $line =~ s/;.+//;
  my @parts = split(/\s*,\s*/, $line);
  if ($#parts <= 0) {
    @parts = split(/\s*<\s*/, $line);
    if ($#parts <= 0) {
      die "invalid_point\n" if $die_on_error;
      emit_warning("Invalid Point");
      return undef;
    }
    ### angular coordinates - figure it out
    my $d    = parse_number($parts[0], $die_on_error);
    my $thed = $parts[1]; # theta
    my $phid = $parts[2] || 0;
    my $the = &Math::Trig::deg2rad($thed);
    my $phi = &Math::Trig::deg2rad($phid);

    my ($x,$y,$z);
    if ($phid == 90 || $phid == 270) {
      $x = 0;
      $y = 0;
      $z = $d;
    } elsif ($thed == 90 || $thed == 270) {
      $x = 0;
      $y = $d * cos($phi);
      $z = $d * sin($phi);
    } elsif ($thed == 0 || $thed == 180) {
      $x = $d * cos($phi);
      $y = 0;
      $z = $d * sin($phi);
    } elsif ($phid == 0 || $phid == 180) {
      $x = $d * cos($the);
      $y = $d * sin($the);
      $z = 0;
    } else {
      $x = $d / sqrt( 1 + sin($phi)**2 / (1 - sin($phi)**2) + sin($the)**2 / (1 - sin($the)**2) );
      $y = $d * sqrt( (1 - sin($phi)**2) / (1 - sin($phi)**2 + (1 - sin($the)**2) / sin($the)**2) );
      $z = $d * sqrt( (1 - sin($the)**2) / (1 - sin($the)**2 + (1 - sin($phi)**2) / sin($phi)**2) );
    }
    @parts = ($x, $y, $z);
    #@parts = &Math::Trig::spherical_to_cartesian($d, $the, $phi);
  } else {
    $_ = parse_number($_, $die_on_error) for @parts;
  }
  if ($is_relative) {
    my @last = @{ last_point() || [] };
    my $max = ($#parts == 2 || $#last == 2) ? 2 : 1;
    $parts[$_] = ($parts[$_] || 0) + ($last[$_] || 0) for 0 .. $max;
  }
  return \@parts;
}

### get a number from the passed in line
sub parse_number {
  my $number = shift;
  my $die_on_error = shift;
  if ($number =~ m/^\s*($QR_NUMBER)\s*($|;)/) {
    return $1;
  ### should handle architectural number entry here
  } else {
    emit_warning("\"$number\" is not a valid number") if ! $die_on_error;
    die "parse_error ($number)\n"; # long jump back up
    $number =~ s/^\s+//;
    $number =~ s/\s+$//;
  }
}

### show options for a command
sub show_usage {
  my $args    = shift;
  my $title   = $args->{title} || '';
  my $pre     = $args->{pre}   || '';
  my $options = $args->{options} || die "Missing options";
  my $default = $args->{default};
  foreach ($title, $pre) {
    $_ = &$_() if ref($_) eq 'CODE';
    s/^\s*//;
    s/\s*$//;
  }
  $title = " $title\n" if $title;
  $pre   = " $pre" if $pre;

  my @opts = ();
  $options = &$options() if ref($options) eq 'CODE';
  foreach (sort keys %$options) {
    next if $options->{$_}->{'no_show'};
    my $val      = lc($_);
    my $shortcut = lc($options->{$_}->{'key'} || '');
    $val =~ s/($shortcut)/\u$1/;
    $val .= " $options->{$_}->{extra}" if $options->{$_}->{extra};
    push @opts, $val;
  }

  my $default_text = '';
  $default = &$default() if ref($default) eq 'CODE';
  if (defined($default)) {
    $default = {option => $default} if ! ref $default;
    state->{default} = $default;
    $default_text = $default->{text} || $default->{option} || '';
  }

  my $msg = "${title}${pre} [" .join("/", @opts). "] <$default_text>: ";
  emit_general($msg);
}

###----------------------------------------------------------------###

sub tab_complete {
  my $line = shift;
  my $cols = shift;

#  dex $line;

  ### complete the command
  my $command;
  my $did_complete;
  if (state->{command}) {
    $command = state->{command};
  } elsif ($line =~ /^(\s*)$/ || $line =~ /^(\s*([^\s;]+)\s*($|;|\s))/) {
    my $len = length($1);
    my $com = defined($2) ? uc($2) : '';
    my @grep = grep {index($_, $com) == 0} sort (keys %COMMANDS, keys %ALIASES);
    return undef if $#grep == -1;
    if ($#grep > 0) {
      foreach (@grep) { $_ .= '*' if $ALIASES{$_} } # show aliases with a *
      return \@grep;
    }
    $command = $grep[0];
    substr($line, 0 , $len, '');
    $did_complete = length($command) != length($com);
  }
  return undef if ! $command;
  return "$command $line" if ! $COMMANDS{$command}->{options} || $did_complete;

  debug $command, $line;

  ### complete options
  my $option;
  if (state->{option}) {
    return undef;
  } elsif ($line =~ /^(\s*)$/ || $line =~ /^(\s*([^\s;]+)\s*($|;|\s))/) {
    my $len  = length($1);
    my $com  = defined($2) ? uc($2) : '';
    my @grep = grep {index($_, $com) == 0} sort keys %{ $COMMANDS{$command}->{options} };
    return "$command $line" if $#grep == -1;
    return \@grep if $#grep > 0;
    $option = $grep[0];
    substr($line, 0 , $len, '');
  }
  return "$command $line" if ! $option;
  return "$command $option $line";
}

sub get_history {
  my $ref = [];

  my $command = state->{command} || '';
  my $option  = state->{option}  || '';
  my $args    = $COMMANDS{$command} || {};
  my $type;
  
  if (! $command) {
    $ref = history()->{'__commands'};
  } elsif (! $option && $args->{options}) {
    $ref = history()->{'__options'}->{$command};
  } else {
    if ($args->{options} && $option) {
      $args = $args->{options}->{uc($option)};
      $args = &$args() if ref($args) eq 'CODE';
    }
    if ($args->{requires}) {
      my $index = state->{highest_found};
      $index = -1 if ! defined($index);

      my $req = $args->{requires};
      $req = &$req() if ref($req) eq 'CODE';
      my $hash = $req->[$index + 1];
      $type = uc($hash->{type} || '');
      $ref = history()->{'__requires'}->{$type} if $type;
    }
  }

  unshift @$ref, $type;
  return $ref;
}

sub save_requires_history {
  my $type  = shift;
  my $value = shift;
  my $ref = history()->{'__requires'}->{$type} ||= [];
  if (! defined($ref->[0]) || $ref->[0] ne $value) {
    pop @$ref if $#$ref == 99;
    unshift @$ref, $value;
  }
}

sub slot_escape_pressed {
  my $state = state;

  if (! $state->{command}) {
    send_prompt();
    return;
  }

  emit_info(sprintf "Cancelled %s via ESC", $state->{command});
  if (state->{is_transparent}) {
    state = $state->{previous_state};
    my $line = '';
    check_requires(state,\$line);

  #} elsif ($state->{option}) {
  #  state = {
  #    command => $state->{command},
  #    alias   => $state->{alias},
  #    context => $state->{context},
  #  };
  #  show_usage();

  } else {
    state = {};
    send_prompt();
  }
}

###----------------------------------------------------------------###

sub run_script {
  my $file = shift;
  active_script = Qt::Process(this);
  active_script->addArgument($file);
  this->connect(active_script, SIGNAL('readyReadStdout()'), SLOT('slot_script_ready_read()'));
  this->connect(active_script, SIGNAL('processExited()'), SLOT('slot_script_exited()'));
  active_script->connect(this, SIGNAL('sig_script_write_stdin(const QString &)'),
                         SLOT('writeToStdin(const QString &)'));
  active_script->start;
}

sub stop_script {
  active_script->closeStdin;
  active_script->tryTerminate();
  active_script = undef;
}

sub slot_script_ready_read {
  while (active_script && active_script->canReadLineStdout) {
    slot_process_line(active_script->readLineStdout);
  }
}

sub slot_script_exited {
  active_script = undef;
}

###----------------------------------------------------------------###

#Command: -layer
# Current layer: "0"
# Enter an option 
# [?/Make/Set/New/ON/OFF/Color/Ltype/LWeight/Plot/PStyle/Freeze/Thaw/LOck/Unlock/stAte]: a
# Enter an option [?/Save/Restore/Edit/Name/Delete/Import/EXport]: 

  ### here we process the line
  # Command: *Cancel*
  # Command: <Switching to: Model>
  # Command: _zoom
  # Specify corner or specify scale factor (nX or n%) or
  # [All/Center/Dynamic/Extents/Previous/Scale/Window] <real time>:
  # Press ESC or ENTER to exit, or right-click to display shortcut menu.
  # *Cancel*
  
1;
