package PerlCAD::Command;

###----------------------------------------------------------------###
#  Copyright 2003 - Paul Seamons                                     #
#  Distributed under the GNU General Public License without warranty #
###----------------------------------------------------------------###

use strict;
use PerlCAD::Dump qw(debug);

###----------------------------------------------------------------###

register_action({
  name      => 'draw_circle_cr',
  menutext  => '&Circle (Center, Radius)',
  tooltip   => 'Draw A Circle by center point and radius',
  whatsthis => '<img src="draw_circle_cr"> <b>Draw Circle</b><br/>Draw a circle with a given center point and radius',
  slot      => 'draw_circle_cr_slot',
  method    => sub {
    this->slot_process_line("CIRCLE CR");
  },
});

register_action({
  name      => 'draw_circle_2p',
  menutext  => '&Circle (2 Points)',
  tooltip   => 'Draw A Circle by two ends of the diameter',
  whatsthis => '<img src="draw_circle_2p"> <b>Draw Circle</b><br/>Draw a circle using two points as the ends of the diameter',
  slot      => 'draw_circle_2p_slot',
  method    => sub {
    this->slot_process_line("CIRCLE 2P");
  },
});

register_action({
  name      => 'draw_circle_3p',
  menutext  => '&Circle (3 Points)',
  tooltip   => 'Draw A Circle by 3 points on the circumference',
  whatsthis => '<img src="draw_circle_3p"> <b>Draw Circle</b><br/>Draw a circle by choosing three points located on the circumference',
  slot      => 'draw_circle_3p_slot',
  method    => sub {
    this->slot_process_line("CIRCLE 3P");
  },
});

###----------------------------------------------------------------###

register_command({
  command => 'CIRCLE',
  aliases => ['C'],
  pre     => 'Specify center OR',
  options => {
    CR => {
      key      => 'C',
      extra    => '(cen radius)',
      no_show  => 1,
      method   => sub { add_circle(&draw_circle_cr) },
      requires => [
        { type => 'POINT',    msg => 'Specify the center point', },
        {
          type => 'DISTANCE',
          msg  => 'Specify a radius (length or second point)',
          default_to_last => 1,
          default_queue   => sub { this->last_point },
          draw => \&draw_sample_circle_cr,
        },
      ],
    },
    TTR      => {
      key    => 'T',
      extra  => '(tan tan rad)',
      method => sub {},
    },
    TTT       => {
      key     => 'T',
      extra   => '(tan tan tan)',
      no_show => 1,
      method  => sub {},
    },
    '3P' => {
      key    => 3,
      method => sub { add_circle(&draw_circle_3p) },
      requires => [{
        type => 'POINT',
        msg  => 'Specify the first point',
      }, {
        type => 'POINT',
        msg  => 'Specify the second point',
        draw => \&draw_circle_2p,
      }, {
        type => 'POINT',
        msg  => 'Specify the third point',
        draw => \&draw_circle_3p,
      }],
    },
    '2P' => {
      key    => 2,
      method => sub { add_circle(&draw_circle_2p) },
      requires => [{
        type => 'POINT',
        msg  => 'Specify the first point of diameter',
      }, {
        type => 'POINT',
        msg  => 'Specify the second point of diameter',
        draw => \&draw_circle_2p,
      }],
    },
  },
  default => {
    activate_on => 'POINT',
    text        => 'center point',
    option      => 'CR',
  },
});

sub add_circle {
  my $rows  = shift;
  my $block = undef;
  my $layer = 'default';
  my ($ltype, $color, $width) = @{ this->parent->get_line_info };
  foreach (@$rows) {
    shift @$_;
    this->parent->doc->db->add_arc($block, $ltype, $layer, $color, $width, @$_);
  }
}

sub draw_sample_circle_cr {
  my $state = shift;
  my $p1    = shift;
  my $p2    = shift; # during cr - this queue holds an unfinished distance
  my $p3    = shift;
  return &draw_circle_cp($state, $p1, $p3);
}

### center / point
sub draw_circle_cp {
  my $state = shift;
  my $p1    = shift;
  my $p2    = shift;
  my $rad = sqrt(  ($p1->[0] - $p2->[0])**2
                 + ($p1->[1] - $p2->[1])**2
                 + ($p1->[2] - $p2->[2])**2);
  return [['arc', $p1, $rad, 0, 360]];
}

### center / radius
sub draw_circle_cr {
  my $state = shift;
  my $p1    = shift;
  my $rad   = shift;
  return [['arc', $p1, $rad, 0, 360]];
}

### 3 points lie on a circle
sub draw_circle_3p {
  my $state = shift;
  my $p1    = shift;
  my $p2    = shift;
  my $p3    = shift;
  my $block = undef;
  my $layer = 'default';

  $_->[2] = 0 foreach $p1, $p2, $p3;
  die "Must be coplanar\n"     if $p1->[2] != $p2->[2] || $p2->[2] != $p3->[2];
  if ($p1->[0] == $p2->[0] && $p1->[1] == $p2->[1]) {
    return draw_circle_2p($p1, $p3);
  } elsif ($p1->[0] == $p3->[0] && $p1->[1] == $p3->[1]) {
    return draw_circle_2p($p1, $p2);
  } elsif ($p2->[0] == $p3->[0] && $p2->[1] == $p3->[1]) {
    return draw_circle_2p($p1, $p2);
  }

  if ($p1->[1] == $p2->[1] && $p2->[1] == $p3->[1]) {
    return draw_circle_2p($p1, $p2) if $p2->[0] == $p3->[0];
    die "Must not be colinear\n";
  }

  if ($p1->[1] == $p2->[1]) {
    ($p2, $p3) = ($p3, $p2);
    if ($p1->[1] == $p2->[1]) {
      return draw_circle_2p($p1, $p2) if $p1->[0] == $p2->[0];
      die "May not be colinear\n";
    }
  } elsif ($p3->[1] == $p2->[1]) {
    ($p1, $p2) = ($p2, $p1);
    if ($p3->[1] == $p2->[1]) {
      return draw_circle_2p($p3, $p2) if $p3->[0] == $p2->[0];
      die "May not be colinear\n";
    }
  }

  #  M1 = - (x2 - x1) / (y2 - y1)                     # slope of perp bisector
  #  B1 = 1/2 (y2 + y1 + (x2**2 - x1**2) / (y2 - y1)) # b of perp bisector
  #  x4 = B2 - B1 / M1 - M2;                          # x intersection
  my $M1 = - ($p2->[0] - $p1->[0]) / ($p2->[1] - $p1->[1]);
  my $M2 = - ($p2->[0] - $p3->[0]) / ($p2->[1] - $p3->[1]);
  my $B1 = 1/2 * ($p2->[1] + $p1->[1] + ($p2->[0] ** 2 - $p1->[0] ** 2) / ($p2->[1] - $p1->[1]));
  my $B2 = 1/2 * ($p2->[1] + $p3->[1] + ($p2->[0] ** 2 - $p3->[0] ** 2) / ($p2->[1] - $p3->[1]));
  die "Points are colinear\n" if $M1 == $M2;
  my $x4 = ($B2 - $B1) / ($M1 - $M2);
  my $y4 = $x4 * $M2 + $B2;
  #die "Did not match ($M1, $B1, $M2, $B2)" if $y4 != $x4 * $M1 + $B1;

  ### show geometric construction
  if (0) {
    my ($ltype, $color, $width) = ();#@{ this->parent->get_line_info };
    foreach my $coord ([$p1->[0], $p1->[1], 0, $p2->[0], $p2->[1], 0],
                       [$p3->[0], $p3->[1], 0, $p2->[0], $p2->[1], 0],
                       [($p1->[0] + $p2->[0])/2, ($p1->[1] + $p2->[1])/2, 0, $x4, $y4, 0],
                       [($p3->[0] + $p2->[0])/2, ($p3->[1] + $p2->[1])/2, 0, $x4, $y4, 0],
                       ) {
      debug $coord;
      this->parent->doc->db->add_line($block, $ltype, $layer, $color, $width, @$coord);
    }
  }

  my $cen = [$x4, $y4, $p1->[2]];
  my $rad = sqrt( ($x4 - $p1->[0]) ** 2 + ($y4 - $p1->[1]) ** 2 );

  return [['arc', $cen, $rad, 0, 360]];
}

### based on endpoints of diameter
sub draw_circle_2p {
  my $state = shift;
  my $p1    = shift;
  my $p2    = shift;
  my $p3    = [];
  $_->[2] ||= 0 for $p1, $p2;
  $p3->[$_] = ($p2->[$_] + $p1->[$_]) / 2 for 0 .. 2;
  my $rad = sqrt(  ($p1->[0] - $p3->[0]) ** 2
                 + ($p1->[1] - $p3->[1]) ** 2
                 + ($p1->[2] - $p3->[2]) ** 2);
  return [['arc', $p3, $rad, 0, 360]];
}

###----------------------------------------------------------------###

1;
