package PerlCAD::CommandBox;

###----------------------------------------------------------------###
#  Copyright 2003 - Paul Seamons                                     #
#  Distributed under the GNU General Public License without warranty #
###----------------------------------------------------------------###

use strict;
use Qt;
use Qt::isa qw(Qt::TextEdit);
use Qt::attributes qw(
                      dev_no
                      prompt
                      history_index
                      history_temp
                      history_active
                      command
                      );
use Qt::signals
  sig_process_line   => ['QString', 'int'],
  sig_escape_pressed => [],
  ;
use Qt::slots
  slot_display_line => ['QString', 'int', 'int'],
  slot_send_prompt  => [],
  slot_press_return => [],
  ;
use constant space_sends_return => 1;

use Data::DumpEx;

sub NEW {
  shift->SUPER::NEW(@_);

  history_index = undef;
  
  setCurrentFont(Qt::Font("Courier", 10)); # need fixed width for some calculations

  setTextFormat(&Qt::PlainText());
  setUndoRedoEnabled(1);
  setTabChangesFocus(0);
  setOverwriteMode(0);
  setReadOnly(0);

  setFocus();
#  grabKeyboard();

  set_prompt('Command:');
  clear();
}

###----------------------------------------------------------------###

sub keyPressEvent {
  my $event = shift;

  if ($event->key == Qt::Key_Escape()) {
    dex "escape pressed";
    emit sig_escape_pressed();
  }

  ### only look at ascii (no backspace - nonascii=0 backspc=8)
  ### unmodified or with shift (shift=8,ctrl=16,alt=32)
  if ($event->ascii > 8 && ($event->state == 0 || $event->state == Qt::ShiftButton())) {

    ### tab completion
    if ($event->key == Qt::Key_Tab()) {
      if ($event->state != Qt::ShiftButton()) {
        tab_completion();
        moveCursor(Qt::TextEdit::MoveEnd(), 0);          
        return;
      } else {
        $event->setState(0);
      }
    } elsif (space_sends_return && $event->key == Qt::Key_Space()) {
      $event = new_return_event();
    }

    moveCursor(Qt::TextEdit::MoveEnd(), 0);
    history_active = undef;

  ### allow for history review
  } elsif ($event->key == Qt::Key_Up()
           || $event->key == Qt::Key_Down()
           || $event->key == Qt::Key_PageUp()
           || $event->key == Qt::Key_PageDown()) {
    review_history($event->key);
    return;

  ### let copy (Ctrl-C) through
  } elsif ($event->key == Qt::Key_Shift()
           || $event->key == Qt::Key_Control()
           || ($event->key == Qt::Key_C() && $event->state == Qt::ControlButton())) {
    # do nothing - allow copy - kind of ugly

  ### send all other keys to the end (unless we are editting the last line)
  } else {
    history_active = undef;
    my ($para, $ind) = (1, 1);
    getCursorPosition($para, $ind);
    if ($para != paragraphs() - 1) {
      moveCursor(Qt::TextEdit::MoveEnd(), 0);
      return;
    } else {
      my $line = text($para);
      if ($line =~ /^([^:]*:\ ?)/ && length($1) >= $ind) {
        moveCursor(Qt::TextEdit::MoveEnd(), 0);
        return;
      }
    }
  }

  ### run the event
  SUPER->keyPressEvent($event);

  ### send off the line
  if ($event->key == Qt::Key_Return()) {
    my $line = text(paragraphs() - 2);
    $line =~ s/^[^:]*://;
    $line =~ s/\s*$//;
    emit sig_process_line($line, dev_no);
    history_active = undef;
  }
}

sub replace_last_line {
  my $item = shift;
  my $para = paragraphs() - 1;
  my $line = text($para);
  if ($line =~ m/^([^:]*:)(.*?)\s*$/) {
    $item = defined($item) ? " $item" : ' ';
    setSelection($para, length($1), $para, length($line), 2);
    setSelectionAttributes(2, Qt::Color(Qt::white()), 0);
    removeSelectedText(2);
    insertAt($item, $para, length($1));
  }
}

sub new_return_event {
  return Qt::KeyEvent(Qt::Event::KeyPress(), Qt::Key_Return(), 10, 0);
}

sub slot_press_return {
  moveCursor(Qt::TextEdit::MoveEnd(), 0);
  my $event = new_return_event();
  keyPressEvent($event);
#  doKeyboarAction(Qt::TextEdit::ActionReturn()); # doesn't work as of perlqt 3.008
}

sub tab_completion {
  my $para  = paragraphs() - 1;
  my $line = text($para);
  chop $line;
  my $copy = $line;
  $copy =~ s/^[^:]*:\s*//;
  my $ret = command->tab_complete($copy);
  if (! defined $ret) {
    # do nothing
  } elsif (ref $ret) {
    my $max = 0;
    foreach (@$ret) { $max = length($_) if length($_) > $max }
    my $met = Qt::FontMetrics(currentFont());
    my $width = $met->width(' ') || 1;
    my $cols = int(wrapColumnOrWidth() / $width);
    my $indent = 3;
    my $space  = $max + 3;
    my $I = int(($cols - $indent) / $space); # columns
    my $J = int(($#$ret + 1) / $I) + 1;     # rows
    my $out  = '';
    foreach my $j (0 .. $J - 1) {
      $out .= ' ' x $indent;
      foreach my $i (0 .. $I - 1) {
        my $val = $ret->[$J * $i + $j];
        last if ! defined $val;
        $out .= (! length($val)) ? "" : ($i == $I - 1) ? $val : sprintf("%-${space}s", $val);
      }
      $out .= "\n";
    }
    $out =~ s/\s+$//;
    append($out);
#    $line =~ s/\s+$/ /;
#    $line .= ' ' if $line =~ /^[^:]+:$/;
    append($line);
  } else {
    replace_last_line($ret);
  }
}

sub review_history {
  my $key  = shift;
  my $ref  = command->get_history;
  my $type = shift(@$ref); # qt can't do list - so pass type as first of ref
  if (! history_active) {
    history_temp  = text(paragraphs() - 1);
    history_temp  =~ s/^[^:]*:\s*//;
    history_temp  =~ s/\s*$//;
    history_index = -2;
  }
  my $item;
  if ($key == Qt::Key_Up()) {
    if (history_index < -1) {
      history_index = 0;
    } elsif (history_active && history_active == -1) {
      history_index += 2;
    }
    $item = (history_index >= $#$ref) ? $ref->[$#$ref] : $ref->[history_index];
    history_index ++;
    history_active = 1;
  } else {
    if (history_index > $#$ref) {
      history_index = $#$ref - 1;
    } elsif (history_active && history_active == 1) {
      history_index -= 2;
    }
    $item = (history_index <= -1) ? history_temp : $ref->[history_index];
    history_index --;
    history_active = -1;
  }
  if (defined $item) {
    replace_last_line($item);
    moveCursor(Qt::TextEdit::MoveEnd(), 0);
    if ($type && $type eq 'POINT') {
      my $point;
      if (history_index <= -1) {
        $point = command->last_point || [0, 0, 0];
      } else {
        $point = [split /\s*,\s*/, $item];
      }
      command->send_relative_point($point);
    }
  }
}

###----------------------------------------------------------------###

sub set_command {
  command = shift;
}

sub set_dev_no {
  dev_no = shift;
}

sub slot_display_line {
  my $msg = shift;
  return if dev_no == shift;
  moveCursor(Qt::TextEdit::MoveEnd(), 0);
#  insertParagraph($msg, paragraphs() - 1);
#  replace_last_line($msg);
#  return;
  append($msg);
  moveCursor(Qt::TextEdit::MoveEnd(), 0);
}

sub slot_send_prompt {
  my $prompt = prompt;
  $prompt =~ s/:?\s*$/: /;
  append($prompt);
  moveCursor(Qt::TextEdit::MoveEnd(), 0);  
}

sub set_prompt {
  my $prompt = shift;
  $prompt =~ s/^\s+//;
  $prompt =~ s/\s+$//;
  prompt = $prompt;
}

sub clear {
  SUPER->clear();
  slot_send_prompt();
}

###----------------------------------------------------------------###

1;
