package PerlCAD::Coord;

###----------------------------------------------------------------###
#  Copyright 2003 - Paul Seamons                                     #
#  Distributed under the GNU General Public License without warranty #
###----------------------------------------------------------------###
use strict;
use warnings;

use Data::DumpEx;

use constant PI => 4 * atan2(1,1); # 3.14159265358979;
use Math::Trig qw(deg2rad rad2deg tan asin);

###----------------------------------------------------------------###

if (__FILE__ eq $0) {
  my $vpoint = __PACKAGE__->new({ vx => 0, vy => 0, vz => 1 });
  dex $vpoint;
  dex $vpoint->project([1,0,0]);
  dex $vpoint->project([1,1,1]);
  dex $vpoint->project([3,2,1]);
}

###----------------------------------------------------------------###

sub new {
  my $class = shift;
  my $self  = bless {}, $class;
  my $args  = shift || {
    vx => 0,
    vy => 0,
    vz => 1,
  };
  $self->set_vpoint([$args->{vx} || 0,
                     $args->{vy} || 0,
                     $args->{vz} || 0,
                     ]);
  return $self;
}

sub is_plan {
  my $self = shift;
  return undef if $self->{vx} != 0 || $self->{vy} != 0 || $self->{vz} == 0;
  return $self->{vz};
}

sub is_side {
  my $self = shift;
  return undef if $self->{vz} != 0 || $self->{vy} != 0 || $self->{vx} == 0;
  return $self->{vx};
}

sub is_elev {
  my $self = shift;
  return undef if $self->{vz} != 0 || $self->{vx} != 0 || $self->{vy} == 0;
  return $self->{vy};
}

sub set_vpoint {
  my $self = shift;
  my ($x, $y, $z) = @{ shift() };
  my $the = atan2($z, $y);
  my $phi = atan2($z, $x);

  $x = $y = $z = 1;

  my $a  = deg2rad(45); # alpha rotation about the x
  my $b  = deg2rad(35.2643896827546); # beta rotation about the y
  my $g  = deg2rad(30); # gamma rotation about the z
  $b = &Math::Trig::asin(&Math::Trig::tan($g));

  my $ca = cos($a);
  my $sa = sin($a);
  my $cb = cos($b);
  my $sb = sin($b);
  my $cg = cos($g);
  my $sg = sin($g);
  
  my $ixr = [[1, 0,    0,   0],
             [0, $ca,  $sa, 0],
             [0, -$sa, $ca, 0],
             [0, 0,    0,   1]]; # inverse rotation x
  
  my $iyr = [[$cb,  0, -$sb, 0],
             [0,    1, 0,    0],
             [$sb,  0, $cb,  0],
             [0,    0, 0,    1]]; # inverse rotation y
  
  my $izr = [[$cg, $sg,   0, 0],
             [-$sg, $cg,  0, 0],
             [0,   0,     1, 0],
             [0,   0,     0, 1]]; # inverse rotation z
  my $matrix = &matrix_mult(&matrix_mult($izr, $iyr), $ixr);


  $self->{vx} = $x;
  $self->{vy} = $y;
  $self->{vz} = $z;
  $self->{vt} = $matrix;
}

sub matrix_mult {
  my $A = shift; # m rows by n columns
  my $B = shift; # n rows by p columns
  my @C = ();    # m rows by p columns

  my $m = $#$A;
  my $p = $#{ $B->[0] };
  my $n = $#{ $A->[0] };
  die "Invalid mult" if $n != $#$B;
  
  for (my $i = 0; $i <= $m; $i ++) {
    for (my $j = 0; $j <= $p; $j ++) {
      my $sum = 0;
      my $t = "($i $j) ";
      for (my $k = 0; $k <= $n; $k ++) {
        $sum += $A->[$i][$k] * $B->[$k][$j];
        $t .= $A->[$i][$k] .'*'. $B->[$k][$j] ." + ";
      }
#      print "$t\n";
      $C[$i][$j] = $sum;
    }
  }
  
  return \@C;
}
#my @A = ([1,  0, 2],
#         [-1, 3, 1]); # n columns by m rows
#my @B = ([3, 1],
#         [2, 1],
#         [1, 0]);     # p columns by n rows
#dex matrix_mult(\@A, \@B);  # should be [[5, 1], [4, 2]];


sub project { 
  my $self = shift;
  my ($x, $y, $z) = @{ shift() };
  my $trans = [[1, 0, 0, 1],
               [0, 1, 0, 1],
               [0, 0, 1, 1],
               [0, 0, 0, 1]]; # inverse object translation
  $trans = &matrix_mult($trans, $self->{vt});
  my $final = &matrix_mult($trans, [[$x], [$y], [$z], [1]]);

#  dex $final;
  #[xprime,                                                  [x,
  # yprmie,    = [ Cameraa transrom] X [World Transform] X    y,
  # zprime,                                                   z,
  # wprime]                                                   1]

  #xfinal = xprime / wprime
  #yfinal = yprime / wprime
  #near = zprime / wprime
  #where -1,-1 is bottom left and 1,1 is top right
  return [$final->[0][0] / $final->[3][0],
          $final->[1][0] / $final->[3][0]];
}

###----------------------------------------------------------------###
1;
__END__

dex point_projection([3,2,1]);
sub PI () {
  return 4 * atan2(1,1); # 3.14159265358979;
}

#1, -1, 1
#-45, 45, -45

sub point_projection {
  my ($x, $y, $z) = @{ shift() };

  ### all object rotations set to zero creates an identity matrix
  if (0) {
    my $obj_trans = [[1, 0, 0, $x],
                     [0, 1, 0, $y],
                     [0, 0, 1, $z],
                     [0, 0, 0, 1 ]]; # object translation
    
    my $a  = -0 * (PI() / 180); # alpha rotation about the x
    my $ca = cos($a);
    my $sa = sin($a);
    my $b  = +0 * (PI() / 180); # beta rotation about the y
    my $cb = cos($b);
    my $sb = sin($b);
    my $g  = -0 * (PI() / 180); # gamma rotation about the z
    my $cg = cos($g);
    my $sg = sin($g);
    
    my $xr = [[1, 0,   0,    0],
              [0, $ca, -$sa, 0],
              [0, $sa, $ca,  0],
              [0, 0,   0,    1]]; # rotation x
    
    my $yr = [[$cb,  0, $sb, 0],
              [0,    1, 0,   0],
              [-$sb, 0, $cb, 0],
              [0,    0, 0,   1]]; # rotation y
    
    my $zr = [[$cg, -$sg, 0, 0],
              [$sg, $cg,  0, 0],
              [0,   0,    1, 0],
              [0,   0,    0, 1]]; # rotation z
    my $tr = &matrix_mult(&matrix_mult($zr, $yr), $xr);
    
    dex $tr;
    
    my $world_transform = &matrix_mult($obj_trans, $tr);
    #[World Transform] = [Translation] X [Rotation]
  }

  my $iobj_trans = [[1, 0, 0, -$x],
                    [0, 1, 0, -$y],
                    [0, 0, 1, -$z],
                    [0, 0, 0, 1  ]]; # inverse object translation

  my $a  = -0 * (PI() / 180); # alpha rotation about the x
  my $ca = cos($a);
  my $sa = sin($a);
  my $b  = +0 * (PI() / 180); # beta rotation about the y
  my $cb = cos($b);
  my $sb = sin($b);
  my $g  = -45 * (PI() / 180); # gamma rotation about the z
  my $cg = cos($g);
  my $sg = sin($g);
  
  my $ixr = [[1, 0,    0,   0],
             [0, $ca,  $sa, 0],
             [0, -$sa, $ca, 0],
             [0, 0,    0,   1]]; # inverse rotation x
  
  my $iyr = [[$cb,  0, -$sb, 0],
             [0,    1, 0,    0],
             [$sb,  0, $cb,  0],
             [0,    0, 0,    1]]; # inverse rotation y
  
  my $izr = [[$cg, $sg,   0, 0],
             [-$sg, $cg,  0, 0],
             [0,   0,     1, 0],
             [0,   0,     0, 1]]; # inverse rotation z
  my $itr = &matrix_mult(&matrix_mult($izr, $iyr), $ixr);
  
  my $camera_transform = &matrix_mult($iobj_trans, $itr);
  #[Camera Transform] = [Inverse ratation] X [Inverse translation]

  ### world was identity
  if (0) {
    #my $sofar = &matrix_mult($camera_transform, $world_transform);
    ##[Transform so far] = [Camera transform] X [World Transform]
    #my $final = &matrix_mult($sofar, [[$x], [$y], [$z], [1]]);
  }

  my $final = &matrix_mult($camera_transform, [[$x], [$y], [$z], [1]]);

  dex $final;
  #[xprime,                                                  [x,
  # yprmie,    = [ Cameraa transrom] X [World Transform] X    y,
  # zprime,                                                   z,
  # wprime]                                                   1]

  #xfinal = xprime / wprime
  #yfinal = yprime / wprime
  #near = zprime / wprime
  #where -1,-1 is bottom left and 1,1 is top right
  return [$final->[0][0] / $final->[3][0],
          $final->[1][0] / $final->[3][0],
          $final->[2][0] / $final->[3][0]];
}

1;
__END__

TreflectXY = [1 0 0  0,
              0 1 0  0,
              0 0 -1 0,
              0 0 0  1];
TreflectX  = [1 0  0  0,
              0 -1 0  0,
              0 0  -1 0,
              0 0  0  1];
Tscaling = [a 0 0 0,
            0 b 0 0,
            0 0 c 0,
            0 0 0 1];
Tplan(XY) = [1 0 0 0,
             0 1 0 0,
             0 0 0 0,
             0 0 0 1];
Telev(XZ) = [0 1 0 0,
             0 0 1 0,
             0 0 0 0,
             0 0 0 1];
Tside(YZ) = [1 0 0 0,
             0 0 1 0,
             0 0 0 0,
             0 0 0 1];
rotate about X by theta
rotate about Y by phi
Tcomposite = [cos(phi) sin(theta)sin(phi) sin(phi)cos(theta) 0,
              0        cos(theta)         -sin(theta)        0,
              0        0                  0                  0,
              0        0                  0                  1];
Pz = [1 0 0 0
      0 1 0 0
      0 0 0 0
      0 0 0 1]
Py = [1 0 0 0
      0 0 0 0
      0 0 1 0
      0 0 0 1]
Px = [0 0 0 0
      0 1 0 0
      0 0 1 0
      0 0 0 1]

