package PerlCAD::DB;

###----------------------------------------------------------------###
#  Copyright 2003 - Paul Seamons                                     #
#  Distributed under the GNU General Public License without warranty #
###----------------------------------------------------------------###

use vars qw($DB_CONNECT_INFO_HOOK
            %LTYPE_MAP
            );
use strict;
use Qt;
use Qt::isa qw(Qt::Object);
use Qt::attributes qw(
                      dbh
                      dev_no
                      filename
                      ids
                      color_by_name
                      ltype_by_name
                      layer_by_name
                      block_by_name
                      ltype_by_id
                      color_by_id
                      layer_by_id
                      no_commit
                      is_bulk_operation
                      last_mod
                      );
use Qt::signals
  sig_process_line => ['QString', 'int'],
  sig_new_data     => ['int', 'bool', 'bool'],
  ;

use DBI;
use POSIX qw(tmpnam);
use constant DEFAULT_LTYPE => 'CONTINUOUS';
use constant DEFAULT_COLOR => 7; # black or white
use Data::DumpEx;


BEGIN {
  %LTYPE_MAP = (  # Qt types - SolidLine, DashLine, DotLine, DashDotLine, DashDotDotLine
    CONTINUOUS   => Qt::SolidLine(),
    HIDDEN       => Qt::DashLine(),
    DASHED       => Qt::DashLine(),
    CENTER       => Qt::DashDotLine(),
    PHANTOM      => Qt::DashDotDotLine(),
    DOT          => Qt::DotLine(),
    DASHDOT      => Qt::DashDotLine(),
    DIVIDE       => Qt::DashDotDotLine(),
    BORDER       => Qt::SolidLine(),
    CONTINUOUS2  => Qt::SolidLine(),
    HIDDEN2      => Qt::DashLine(),
    DASHED2      => Qt::DashLine(),
    CENTER2      => Qt::DashDotLine(),
    PHANTOM2     => Qt::DashDotDotLine(),
    DOT2         => Qt::DotLine(),
    DASHDOT2     => Qt::DashDotLine(),
    DIVIDE2      => Qt::DashDotDotLine(),
    BORDER2      => Qt::SolidLine(),
    CONTINUOUSX2 => Qt::SolidLine(),
    HIDDENX2     => Qt::DashLine(),
    DASHEDX2     => Qt::DashLine(),
    CENTERX2     => Qt::DashDotLine(),
    PHANTOMX2    => Qt::DashDotDotLine(),
    DOTX2        => Qt::DotLine(),
    DASHDOTX2    => Qt::DashDotLine(),
    DIVIDEX2     => Qt::DashDotDotLine(),
    BORDERX2     => Qt::SolidLine(),
  );
}

sub NEW {
  my $class = shift;
  my $app   = shift;
  my $name  = shift;

  $class->SUPER::NEW($app, $name);

  ids = {
    ltype => 1,
    color => 1,
    layer => 1,
    block => 1,
    line  => 1,
    arc   => 1,
  };
  ltype_by_name = {};
  color_by_name = {};
  layer_by_name = {};
  block_by_name = {};
  ltype_by_id = {};
  color_by_id = {};
  layer_by_id = {};

  set_no_commit(1);
  last_mod = 0;
}

sub commit {
  dbh->commit;
}

sub set_no_commit {
  no_commit = shift;
  commit if ! no_commit;
}

sub set_is_bulk_operation {
  is_bulk_operation = shift;
  if (! is_bulk_operation) {
    emit sig_new_data(last_mod, 1, 1);
  }
}

sub initialize_storage {
  my $document_filename = shift;
  filename = $document_filename;
  filename = "/tmp/tmpsql";

  ### allow for getting the info
  my @info;
  if ($DB_CONNECT_INFO_HOOK && ref($DB_CONNECT_INFO_HOOK) eq 'CODE') {
    @info = &$DB_CONNECT_INFO_HOOK($document_filename);

  ### or use the standard SQLite
  } else {

    ### remove files that are in the way
    my $journal = filename . '-journal'; # must remove the journal as well (if any)
    foreach (filename, $journal) {
      unlink $_ if -e $_;
    }
    
    @info = ("dbi:SQLite:dbname=".filename, "", "", {AutoCommit => 0, RaiseError => 1});
  }

  ### now instantiate a fresh one
  dbh = DBI->connect(@info);
  if (! dbh) {
    dex "$DBI::errstr";
    die $DBI::errstr;
  }

  ### load the tables
  instantiate_schema();
}

###----------------------------------------------------------------###

sub set_dev_no {
  dev_no = shift;
}

sub get_dev_no {
  return dev_no;
}

sub send_process_line {
  my $line = shift;
  emit sig_process_line($line, dev_no);
}

###----------------------------------------------------------------###

sub add_comment {
  ids->{comment} ||= 0;
  ids->{comment} ++;
  dbh->do("INSERT INTO comment
             (comment_id, comment)
           VALUES (?, ?)", {}, ids->{comment}, @_);
  dbh->commit if ! no_commit;
}

sub add_variable {
  ids->{variable} ||= 0;
  ids->{variable} ++;
  dbh->do("INSERT INTO variable
             (var_id, key, value, section)
           VALUES (?, ?, ?, ?)", {}, ids->{variable}, @_);
  dbh->commit if ! no_commit;
}

###----------------------------------------------------------------###

sub add_ltype {
  my $id   = ids->{ltype} ++;
  my $name = shift;
  my $val  = $LTYPE_MAP{uc($name)} || Qt::SolidLine();

  dbh->do("INSERT INTO ltype
             (ltype_id, name, qt_value, desc, align, n_items, length, dashes)
           VALUES (?,   ?,    ?,        ?,    ?,     ?,       ?,      ?)", {},
          $id, $name, $val, @_);
  dbh->commit if ! no_commit;
  ltype_by_name->{$name} = $id;
  return $id;
}

sub add_color {
  my $id   = ids->{color} ++;
  my $name = shift;
  $name = DEFAULT_COLOR if ! defined $name;
  dbh->do("INSERT INTO color
             (color_id, name)
           VALUES (?, ?)", {}, $id, $name, @_);
  dbh->commit if ! no_commit;
  color_by_name->{$name} = $id;
  return $id;
}

sub add_layer {
  my $id     = ids->{layer} ++;
  my $name   = shift;
  my $ltype  = shift || DEFAULT_LTYPE;
  my $color  = shift || DEFAULT_COLOR;
  my $width  = shift || 0;
  my $hidden = shift || 0;
  my $locked = shift || 0;
  my $ltype_id = get_ltype_id_by_name($ltype);
  my $color_id = get_color_id_by_name($color);
  dbh->do("INSERT INTO layer
             (layer_id, name, ltype_id, color_id, width, hidden, locked, flags, sdump)
           VALUES (?,   ?,    ?,        ?,        ?,     ?,      ?,      ?,     ?)", {},
          $id, $name, $ltype_id, $color_id, $width, $hidden, $locked, shift(), shift());
  dbh->commit if ! no_commit;
  layer_by_name->{$name} = $id;
  return $id;
}

sub add_block {
  my $id = ids->{block} ++;
  die "Usage add_block" if $#_ != 8 - 1;
  my ($type, $name, $desc, $layer, $x, $y, $z, $dumpx) = @_;
  my $layer_id = get_layer_id_by_name($layer || 'default');
  $_ ||= 0 foreach $x, $y, $z;
  dbh->do("INSERT INTO block
             (block_id, name, description, type, layer_id, x, y, z, dump)
           VALUES (?,   ?,    ?,           ?,    ?,        ?, ?, ?, ?)", {},
          $id, $name, $desc, $type, $layer_id, $x, $y, $z, $dumpx);
  dbh->commit if ! no_commit;
  block_by_name->{$name} = $id;
  return $id;
}

###

sub get_ltype_id_by_name {
  my $name = shift;
  $name = 'ByLayer' if ! defined $name;
  if (! exists ltype_by_name->{$name}) {
    my $s = "SELECT ltype_id, name FROM ltype";
    my $rows = dbh->selectall_arrayref($s);
    ltype_by_name->{$_->[1]} = $_->[0] foreach @$rows;
    add_ltype($name, 'autoadd','','','','') if ! exists ltype_by_name->{$name};
  }
  return ltype_by_name->{$name};
}

sub get_color_id_by_name {
  my $name = shift;
  $name = 'ByLayer' if ! defined $name;
  if (! exists color_by_name->{$name}) {
    my $s = "SELECT color_id, name FROM color";
    my $rows = dbh->selectall_arrayref($s);
    color_by_name->{$_->[1]} = $_->[0] foreach @$rows;
    add_color($name) if ! exists color_by_name->{$name};
  }
  return color_by_name->{$name};
}

sub get_layer_id_by_name {
  my $name = shift;
  $name = 'default' if ! defined $name;
  if (! exists layer_by_name->{$name}) {
    my $s = "SELECT layer_id, name FROM layer";
    my $rows = dbh->selectall_arrayref($s);
    layer_by_name->{$_->[1]} = $_->[0] foreach @$rows;
    add_layer($name, DEFAULT_LTYPE, DEFAULT_COLOR, 0, 0, 0, 0, 'autoadd') if ! exists layer_by_name->{$name};
  }
  return layer_by_name->{$name};
}

sub get_block_id_by_name {
  my $name = shift;
  if (! exists block_by_name->{$name}) {
    my $s = "SELECT block_id, name FROM block";
    my $rows = dbh->selectall_arrayref($s);
    block_by_name->{$_->[1]} = $_->[0] foreach @$rows;
    if (! exists block_by_name->{$name}) {
      die "No block by name $name" if uc($name) ne 'BYLAYER';
      add_block('block', $name, 'autoadd', 0, 0, 0, 0, '');
    }
  }
  return block_by_name->{$name};
}

sub get_qtcolor_by_id {
  my $id = shift;
  my $color = color_by_id ||= do {
    my $s = "SELECT color_id, qt_value FROM color";
    my $rows = dbh->selectall_arrayref($s);
    my %hash = map {$_->[0] => $_->[1]} @$rows;
    \%hash; # return of the do
  };
  return $color->{$id} ||= DEFAULT_COLOR;
}

sub get_penstyle_by_id {
  my $id = shift;
  my $styles = ltype_by_id ||= do {
    my $s = "SELECT ltype_id, qt_value FROM ltype";
    my $rows = dbh->selectall_arrayref($s);
    my %hash = map {$_->[0] => $_->[1]} @$rows;
    \%hash; # return of the do
  };
  return $styles->{$id} ||= Qt::SolidLine();
}

###----------------------------------------------------------------###

sub add_line {
  my $line_id = ids->{line} ++;
  die "Usage add_line" if $#_ != 7 - 1;
  my ($block, $ltype, $layer, $color, $width, $p1, $p2) = @_;
  my ($x1, $y1, $z1) = @$p1;
  my ($x2, $y2, $z2) = @$p2;
  my $block_id = $block ? get_block_id_by_name($block) : 0;
  my $ltype_id = get_ltype_id_by_name($ltype);
  my $color_id = get_color_id_by_name($color);
  my $layer_id = get_layer_id_by_name($layer);
  $_ ||= 0 foreach $x1, $y1, $z1, $x2, $y2, $z2, $width;
  dbh->do("INSERT INTO line
             (line_id,  modified, selected, block_id, ltype_id, color_id, layer_id, width, x1, y1, z1, x2, y2, z2)
           VALUES (?,   ?,        ?,        ?,        ?,        ?,        ?,        ?,     ?,  ?,  ?,  ?,  ?,  ?)", {},
          $line_id, ++last_mod, 0, $block_id, $ltype_id, $color_id, $layer_id, $width, $x1, $y1, $z1, $x2, $y2, $z2);
  dbh->commit if ! no_commit;
  if (! is_bulk_operation) {
    emit sig_new_data(last_mod, 1, 0);
  }
  return $line_id;
}

sub add_arc {
  my $arc_id = ids->{arc} ++;
  die "Usage add_arc" if $#_ != 9 - 1;
  my ($block, $ltype, $layer, $color, $width, $p1, $radius, $ang1, $ang2) = @_;
  my ($x1, $y1, $z1) = @$p1;
  my $block_id = $block ? get_block_id_by_name($block) : 0;
  my $ltype_id = get_ltype_id_by_name($ltype);
  my $color_id = get_color_id_by_name($color);
  my $layer_id = get_layer_id_by_name($layer);
  $_ ||= 0 foreach $x1, $y1, $z1, $radius, $ang1, $ang2, $width;
  dbh->do("INSERT INTO arc
             (arc_id,  modified, selected, block_id,  ltype_id,  color_id,  layer_id,  width,
              x1,  y1,  z1,  radius, angle1, angle2)
           VALUES (?,  ?,        ?,        ?,         ?,         ?,         ?,         ?,
              ?,   ?,   ?,   ?,      ?,      ?)", {},
          $arc_id, ++last_mod, 0, $block_id, $ltype_id, $color_id, $layer_id, $width,
          $x1, $y1, $z1, $radius, $ang1, $ang2);
  dbh->commit if ! no_commit;
  if (! is_bulk_operation) {
    emit sig_new_data(last_mod, 0, 1);
  }
  return $arc_id;
}

###----------------------------------------------------------------###

sub instantiate_schema {
  my $ref = schema();
  for my $i (0 .. int($#$ref / 2)) {
    my $table  = $ref->[2 * $i];
    my $fields = $ref->[2 * $i + 1];
    my $sql = "CREATE TABLE $table (\n";
    $sql .= join ",\n", map {$fields->[2 * $_] ." ". $fields->[2 * $_ + 1]} 0 .. int($#$fields / 2);
    $sql .= "\n)";
#    dex $sql;
    dbh->do($sql);
  }

  this->create_indexes;
}

sub create_indexes {
}

sub schema {
  [
   'comment' => [
                 'comment_id' => 'INTEGER PRIMARY KEY',
                 'comment'    => 'VARCHAR(1000)',
                 ],
   'variable' => [
                  'var_id'  => 'INTEGER PRIMARY KEY',
                  'key'     => 'VARCHAR(100) NOT NULL',
                  'value'   => 'VARCHAR(100)',
                  'section' => 'VARCHAR(20)',
                  ],
   'ltype' => [
               'ltype_id' => 'INTEGER PRIMARY KEY',
               'name'     => 'VARCHAR(100) NOT NULL',
               'qt_value' => 'INTEGER NOT NULL',
               'desc'     => 'VARCHAR(100)',
               'align'    => 'INTEGER',
               'n_items'  => 'INTEGER',
               'length'   => 'NUMBER',
               'dashes'   => 'VARCHAR(50)',
               ],
   'color' => [
               'color_id' => 'INTEGER PRIMARY KEY',
               'name'     => 'VARCHAR(100)',
               ],
   'layer' => [
               'layer_id' => 'INTEGER PRIMARY KEY',
               'name'     => 'VARCHAR(100) NOT NULL',
               'ltype_id' => 'VARCHAR(100) NOT NULL',
               'color_id' => 'INTEGER',
               'width'    => 'VARCHAR(15)',
               'hidden'   => 'INTEGER',
               'locked'   => 'INTEGER',
               'flags'    => 'INTEGER',
               'sdump'    => 'VARCHAR(1000)',
               ],
   'block' => [
               'block_id'    => 'INTEGER PRIMARY KEY',
               'name'        => 'VARCHAR(100) NOT NULL',
               'description' => 'VARCHAR(100) NOT NULL',
               'type'        => 'VARCHAR(10) NOT NULL', # (block, xref, poly)
               'layer_id'    => 'INTEGER NOT NULL',
               'x'           => 'NUMBER NOT NULL',
               'y'           => 'NUMBER NOT NULL',
               'z'           => 'NUMBER NOT NULL',
               'dump'        => 'VARCHAR(1000)',
               ],
   'line' => [
              'line_id'  => 'INTEGER PRIMARY KEY',
              'block_id' => 'INTEGER NOT NULL',
              'ltype_id' => 'INTEGER NOT NULL',
              'layer_id' => 'INTEGER NOT NULL',
              'color_id' => 'INTEGER NOT NULL',
              'width'    => 'NUMBER NOT NULL',
              'x1'       => 'NUMBER NOT NULL',
              'y1'       => 'NUMBER NOT NULL',
              'z1'       => 'NUMBER NOT NULL',
              'x2'       => 'NUMBER NOT NULL',
              'y2'       => 'NUMBER NOT NULL',
              'z2'       => 'NUMBER NOT NULL',
              'modified' => 'INTEGER NOT NULL',
              'selected' => 'INTEGER NOT NULL',
              ],
   'arc' => [
             'arc_id'   => 'INTEGER PRIMARY KEY',
             'block_id' => 'INTEGER NOT NULL',
             'ltype_id' => 'INTEGER NOT NULL',
             'layer_id' => 'INTEGER NOT NULL',
             'color_id' => 'INTEGER NOT NULL',
             'width'    => 'NUMBER NOT NULL',
             'x1'       => 'NUMBER NOT NULL',
             'y1'       => 'NUMBER NOT NULL',
             'z1'       => 'NUMBER NOT NULL',
             'radius'   => 'NUMBER NOT NULL',
             'angle1'   => 'NUMBER NOT NULL',
             'angle2'   => 'NUMBER NOT NULL',
             'modified' => 'INTEGER NOT NULL',
             'selected' => 'INTEGER NOT NULL',
             ],
   ];
}

###----------------------------------------------------------------###

1;
