package PerlCAD::Doc;

###----------------------------------------------------------------###
#  Copyright 2003 - Paul Seamons                                     #
#  Distributed under the GNU General Public License without warranty #
###----------------------------------------------------------------###

use strict;
use Qt;
use Qt::isa qw(Qt::Object);
use Qt::attributes qw(
                      views
                      active_view
                      filename
                      io
                      db
                      modified
                      modified_confirm
		      last_queried
                      read_only
                      );
use Qt::signals
  toggle_modified => ['int'],
  change_filename => ['QString'],
  sig_new_objects => ['int'],
  ;

use PerlCAD::DB;
use PerlCAD::IO;

use Data::DumpEx;

sub NEW {
  my $class = shift;
  my $app  = shift;
  my $name = shift;
  $class->SUPER::NEW($app, $name);

  io = PerlCAD::IO->new(); # non-Qt object

  set_modified(0);
  set_filename('');
  last_queried = time;

  views = [];
}

###----------------------------------------------------------------###

sub set_modified {
  modified = shift || 0;
  modified_confirm = undef;
  emit toggle_modified(modified);
}

sub set_filename {
  filename = shift || "";
  emit change_filename(filename);
}

sub is_modified {
  return modified;
}

sub get_filename {
  return (filename && filename =~ m|([^/\\]+)$|) ? $1 : 'Untitled';
}

sub is_read_only {
  return read_only;
}

###----------------------------------------------------------------###

sub add_view {
  my $views = views;
  my $view  = shift;
  push @$views, $view;
  $view->connect(db, SIGNAL("sig_new_data(int, bool, bool)"),
                 SLOT("slot_new_data(int, bool, bool)")) if db;
}

sub remove_view {
  my $views = views;
  my $view  = shift;

  ### remove the last added if none passed
  $view ||= $views->[$#$views];
    
  foreach my $i (0 .. $#$views) {
    next if $view ne $views->[$i];
    my $_view = (splice(@$views, $i, 1, ()))[0];
    this->parent->command->unregister_device($_view);
    active_view = undef if $_view eq active_view;
    $_view->disconnect(db, SIGNAL("sig_new_data(int, bool, bool)")) if db;
    last;
  }
  return undef;
}

sub get_views {
  return views;
}

sub count_views {
  return $#{ views() } + 1;
}

sub set_active_view {
  my $view = shift;
  if (active_view) {
    return if $view eq active_view;
    active_view->set_is_active_view(undef);
  }
  active_view = $view;
}

sub get_active_view {
  return active_view || do {
    die "Active view called without an active view";
  };
}

###----------------------------------------------------------------###

sub import_filters {
  return io->import_filters();
}

sub export_filters {
  return io->export_filters();
}

sub import_filters_as_strings {
  return filters_as_strings(import_filters());
}

sub export_filters_as_strings {
  return filters_as_strings(export_filters());
}

sub filters_as_strings {
  my $filters = shift || return '';
  my @array   = ();
  foreach my $row (@$filters) {
    push @array, "$row->[1] (*.$row->[0])";
  }
  return \@array;
}

###----------------------------------------------------------------###

sub sql_line {
  return "SELECT l.line_id, l.block_id, lt.qt_value, c.name, l.layer_id, l.width,
                 l.x1, l.y1, l.z1, l.x2, l.y2, l.z2
            FROM line l, ltype lt, color c, layer la
           WHERE l.ltype_id = lt.ltype_id
             AND l.color_id = c.color_id
             AND l.modified >= ?
             AND l.modified <= ?";
}

sub sql_arc {
  return "SELECT a.arc_id, a.block_id, lt.qt_value, c.name, a.layer_id, a.width,
                 a.x1, a.y1, a.z1, a.radius, a.angle1, a.angle2
            FROM arc a, ltype lt, color c
           WHERE a.ltype_id = lt.ltype_id
             AND a.color_id = c.color_id
             AND a.modified >= ?
             AND a.modified <= ?";
}

### return cursor for lines
sub query_lines {
  my ($start, $end) = @_;
  my @rows = ();
  return if ! db || ! db->dbh;

  ### get normal lines from the database
  my $sth = db->dbh->prepare(sql_line());
  $sth->execute($start, $end);
  my $db = db;
  return sub {
    my $row = $sth->fetchrow_arrayref;
    return $row;
  };
}

### return cursor for arcs
sub query_arcs {
  my ($start, $end) = @_;
  my @rows = ();
  return if ! db || ! db->dbh;
  
  ### get normal lines from the database
  my $sth = db->dbh->prepare(sql_arc());
  $sth->execute($start, $end);
  my $db = db;
  return sub {
    my $row = $sth->fetchrow_arrayref;
    return $row;
  };
}

sub query_layer {
  my $name = shift;
  my $s = "SELECT la.layer_id, lt.qt_value, c.name, la.width, la.hidden, la.locked
             FROM layer la, ltype lt, color c
            WHERE la.ltype_id = lt.ltype_id
              AND la.color_id = c.color_id
              AND LOWER(la.name) = ?";
  my $row = [];
  if (db && db->dbh) {
    $row = db->dbh->selectrow_arrayref($s, {}, lc($name));
  }
  $row->[0] ||= 0;
  $row->[1] ||= $PerlCAD::DB::DEFAULT_LTYPE;
  $row->[2] ||= $PerlCAD::DB::DEFAULT_COLOR;
  $row->[3] ||= 0;
  $row->[4] ||= 0;
  $row->[5] ||= 0;
  return $row;
}

sub query_last_mod {
  return 0 if ! db;
  return db->last_mod || 0;
}

###----------------------------------------------------------------###

sub message_box {
  my $message = shift;
  my $level   = shift || '';
  my $caption = shift || '';
  my $button1 = shift || &Qt::MessageBox::Ok() | &Qt::MessageBox::Default();
  my $button2 = shift || &Qt::MessageBox::NoButton();
  my $button3 = shift || &Qt::MessageBox::NoButton();
  my $icon = 
      ($level eq 'warn')     ? &Qt::MessageBox::Warning()
    : ($level eq 'info')     ? &Qt::MessageBox::Information()
    : ($level eq 'critical') ? &Qt::MessageBox::Critical()
    : &Qt::MessageBox::NoIcon();
  return Qt::MessageBox($caption,
                        $message,
                        $icon,
                        $button1,
                        $button2,
                        $button3,
                        this->parent,
                        'message dialog',
                        1)->exec;
}

###----------------------------------------------------------------###

sub initialize_db {
  my $filename = shift;
  db->dbh->disconnect if db && db->dbh;
  db = PerlCAD::DB(this->parent, "Entity DB Storage");
  db->initialize_storage($filename);
  foreach my $view (@{ &get_views() }) {
    $view->connect(db, SIGNAL("sig_new_data(int, bool, bool)"),
                   SLOT("slot_new_data(int, bool, bool)"));
  }
}

sub new_document {
  if (modified) {
    save_modified() || return;
  }
  filename = undef;

  ### set a fresh dbh
  initialize_db('unamed.pcsql');

  ### test document
  db->add_line(undef, 'CONTINUOUS', 'default', 'Blue',  0,  [2,  2, 2],  [2, 4, 2 ]);
  db->add_line(undef, 'CONTINUOUS', 'default', 'Blue',  0,  [2,  4, 2],  [4, 4, 2 ]);
  db->add_line(undef, 'CONTINUOUS', 'default', 'Blue',  0,  [4,  4, 2],  [4, 2, 2 ]);
  db->add_line(undef, 'CONTINUOUS', 'default', 'Blue',  0,  [4,  2, 2],  [2, 2, 2 ]);

  db->add_line(undef, 'CONTINUOUS', 'default', 'red',    0,  [2,  2, 4],  [2, 4, 4 ]);
  db->add_line(undef, 'CONTINUOUS', 'default', 'red',    0,  [2,  4, 4],  [4, 4, 4 ]);
  db->add_line(undef, 'CONTINUOUS', 'default', 'red',    0,  [4,  4, 4],  [4, 2, 4 ]);
  db->add_line(undef, 'CONTINUOUS', 'default', 'red',    0,  [4,  2, 4],  [2, 2, 4 ]);

  db->add_line(undef, 'CONTINUOUS', 'default', 'green',  0,  [2,  2, 2],  [2, 2, 4 ]);
  db->add_line(undef, 'CONTINUOUS', 'default', 'green',  0,  [4,  2, 2],  [4, 2, 4 ]);
  db->add_line(undef, 'CONTINUOUS', 'default', 'yellow', 0,  [2,  4, 2],  [2, 4, 4 ]);
  db->add_line(undef, 'CONTINUOUS', 'default', 'yellow', 0,  [4,  4, 2],  [4, 4, 4 ]);

#  db->add_line(undef, 'CONTINUOUS', 'default', 'Blue',  0,  [2,  1, 10],  [15, 4, 0 ]);
#  db->add_line(undef, 'CONTINUOUS', 'default', 'Blue',  0,  [15, 4, 0 ],  [12, 8, 5 ]);
#  db->add_line(undef, 'DASHED',     'default', 'Green', 0,  [12, 8, 5 ],  [3,  7, 0 ]);
#  db->add_line(undef, 'DOT',        'default', 'Blue',  1,  [3,  7, 0 ],  [2,  8, 3 ]);
#  db->add_line(undef, 'CONTINUOUS', 'default', 'Red',   0,  [2,  8, 3 ],  [2,  1, 10]);
  db->commit;

  set_modified(0);
  set_filename('');

  this->parent->prop_layer->set_layername('default');

  ### allow for a screen reload
  foreach my $view (@{ views() }) {
    $view->reload();
    $view->zoom_extents();
  }
  return 1;
}

sub open_document_nondialog_requires {
  my $command = shift;
  my $state   = shift;
  my $lineref = shift;
  
  if (modified && ! modified_confirm) {
    my $ok = 0;
    if ($$lineref =~ s/$PerlCAD::Command::QR_CHUNK_FROM_LINE//) {
      my $ans = lc($1);
      if ($ans eq 'y' || $ans eq 'yes') {
        warn "Unimplemented save";
        $ok = 1;
      } elsif ($ans eq 'n' || $ans eq 'no') {
        # do nothing
        $ok = 1;
        modified_confirm = 1; # has bug of not resetting on cancelled command
      } else {
      }
    }
    if (! $ok) {
      return [{type => 'CUSTOM', msg => 'Save current document? (Y/N)', method => sub { 0 }}];
    }
  }

  return [{type => 'FILENAME', msg => 'Specify filename to open'}];
}

sub open_document {
  ### prompt for a filename if none
  my $filename = shift;
  my $state    = shift;

  ### if old document is modified - save it
  if (modified && (! $state || $state->{use_dialog})) {
    save_modified() || return;
  }
  
  my $filters = import_filters_as_strings();
  my $filter;
  my $filter_index;

  ### query for it
  if (! $filename) {
    my $d = Qt::FileDialog(
                           "/home/pauls/cad/dxf",
                           $filters->[0],
                           this->parent,
                           "open file dialog",
                           );
    $d->setCaption("Open file...");
    $d->addFilter($filters->[$_]) for 1 .. $#$filters; # add the rest of the filters

    my $status = $d->exec();
    return undef if ! defined($status) || $status != Qt::Dialog::Accepted();

    $filename = $d->selectedFile();
    $filter   = $d->selectedFilter();
    for (0 .. $#$filters) {
      next if $filters->[$_] ne $filter;
      $filter_index = $_;
      last;
    }
    return if ! defined($filename) || ! length($filename);
    return if ! defined $filter_index;
  }

  ### test the filename and open it
  my $fh;
  my $size = -s $filename;
  if (! -e _) {
    message_box("File $filename could not be found", 'critical');
    return;
  } elsif (! $size) {
    my $r = message_box("File $filename is zero bytes - load anyway?", 'warn', '',
                        &Qt::MessageBox::Yes() | &Qt::MessageBox::Default(),
                        &Qt::MessageBox::No()  | &Qt::MessageBox::Escape(),
                        );
    return if $r != &Qt::MessageBox::Yes();
  }
  if (! open($fh, "<", $filename)) {
    message_box("Couldn't open $filename: $!", 'critical');
    return;
  }

  ### set a fresh dbh
  initialize_db($filename);

  ### now import the file
  ### register the device temporarily so we can tell all devices what is going on
  ### this also allows for blocking of any other menu options so that
  ### this can be modal
  if ($fh && $size) {

    ### allow for taking control of the command line
    my $command = this->parent->get_command;
    $command->register_device(db, 'sig_process_line', undef);
    $command->set_dedicated(db);

    ### get a progress dialog
    my $dia = Qt::ProgressDialog("Loading file...", "Cancel", $size, this->parent, "file import progress dialog", 1);
    $dia->setAutoClose(0);
    $dia->setMinimumDuration(4000);

    my $db = db;
    $db->set_is_bulk_operation(1);

    ### do the import
    my $last = 0;
    io->register_hooks({
      debug         => sub { $db->send_process_line(@_) },

      data_progress => sub {
        my $read  = shift;
        my $total = shift;
        $last = int($read / $total * 100);
#        $DB->commit if ! ($last % 20);
        $dia->setProgress($read, $total) if $last != 100;
#      $dia->parent->parent->processEvents();
        
      },
      add_comment   => sub { $db->add_comment(@_)  },
      add_variable  => sub { $db->add_variable(@_) },
      add_block     => sub { $db->add_block(@_)    },
      add_ltype     => sub { $db->add_ltype(@_)    },
      add_layer     => sub { $db->add_layer(@_)    },
      add_line      => sub { $db->add_line(@_)     },
      add_arc       => sub { $db->add_arc(@_)      },
      
    });
    io->run_import($filter_index, $fh, $size);
    io->clear_hooks();

    close($fh);

    db->commit();
    $db->set_is_bulk_operation(undef);

    ### close progress dialog
    $dia->setAutoClose(1);
    $dia->reset();

    ### un register from command
    $command->unset_dedicated;
    $command->unregister_device(db);
  }

  set_modified(1);
  set_filename($filename);

  ### allow for a screen reload
  foreach my $view (@{ views() }) {
    $view->reload();
    $view->zoom_extents();
  }

  return 1;
}

sub save_document {
  my $filename = shift || filename;

  return save_document_as() if ! $filename;

  db->commit;
  dex "Should save filename $filename now";

  set_modified(0);
  set_filename($filename);

  return 1;
}

sub save_document_as {
  my $filename = shift
    || Qt::FileDialog::getSaveFileName(
                                       Qt::Dir::currentDirPath(),
                                       export_filters_as_string(),
                                       this->parent,
                                       "save as file dialog",
                                       "Save file as...",
                                       );
  if (! $filename) {
    return 0;
  } elsif (-e $filename) {
    my $r = message_box("The file: $filename already exists.  Overwrite?",
                        'warn', '',
                        &Qt::MessageBox::Yes()    | &Qt::MessageBox::Default(),
                        &Qt::MessageBox::No(),
                        &Qt::MessageBox::Cancel() | &Qt::MessageBox::Escape(),
                        );
    return 0 if $r != &Qt::MessageBox::Yes();
  }

  return save_document($filename);
}

sub save_modified {
  return 1 if ! modified;

  my $r = message_box("The current file has been modified.\nDo you want to save it?",
                      'warn', '',
                      &Qt::MessageBox::Yes()    | &Qt::MessageBox::Default(),
                      &Qt::MessageBox::No(),
                      &Qt::MessageBox::Cancel() | &Qt::MessageBox::Escape(),
                      );
  if ($r == &Qt::MessageBox::Yes()) {
    if (! filename || filename =~ m|[/\\]Untitled$|) {
      return save_document_as();
    } else {
      return save_document();
    }
  } elsif ($r == &Qt::MessageBox::No) {
    # don't do anything - they could still cancel on an open
    return 1;
  }
  return 0;
}

###----------------------------------------------------------------###

sub get_extents {
  return [0, 0, 0, 12, 9, 0] if ! db || ! db->dbh;
  my $s = "SELECT MIN(MIN(x1,x2)), MIN(MIN(y1,y2)), MIN(MIN(z1,z2)) FROM line";
  my ($minx, $miny, $minz) = db->dbh->selectrow_array($s);
     $s = "SELECT MAX(MAX(x1,x2)), MAX(MAX(y1,y2)), MAX(MAX(z1,z2)) FROM line";
  my ($maxx, $maxy, $maxz) = db->dbh->selectrow_array($s);
  $_ ||= 0 foreach $minx, $miny, $minz, $maxx, $maxy, $maxz;
  $maxx += 1 if $minx == $maxx;
  $maxy += 1 if $miny == $maxy;
  return [$minx, $miny, $minz, $maxx, $maxy, $maxz];
}

###----------------------------------------------------------------###

1;
