package PerlCAD::IO;

###----------------------------------------------------------------###
#  Copyright 2004 - Paul Seamons                                     #
#  Distributed under the GNU General Public License without warranty #
###----------------------------------------------------------------###

use strict;
use Data::DumpEx;

use vars qw(@NAMESPACE
            @IMPORT_FILTERS
            @EXPORT_FILTERS
            $loaded
            %method_map
            $debug_required
            );

BEGIN {
  ### list of namespaces to use when looking for IO filters
  @NAMESPACE = (__PACKAGE__);
};

use constant _debug         => 5;
use constant _data_progress => 6;
use constant _add_comment   => 10;
use constant _add_variable  => 11;
use constant _add_block     => 12;
use constant _add_ltype     => 13;
use constant _add_layer     => 14;
use constant _add_line      => 15;
use constant _add_arc       => 16;

###----------------------------------------------------------------###
### public member functions

sub hook_map {
  return {
    'debug'         => _debug,
    'data_progress' => _data_progress,
    'add_comment'   => _add_comment,
    'add_variable'  => _add_variable,
    'add_block'     => _add_block,
    'add_ltype'     => _add_ltype,
    'add_layer'     => _add_layer,
    'add_line'      => _add_line,
    'add_arc'       => _add_arc,
  };
}

sub register_io_import {
  my $args = shift || die "Missing arguments";
  my $ext  = lc($args->{extension}) || die "Missing extension";
  my $desc = $args->{description}   || die "Missing description";
  my $func = $args->{handler}       || die "Missing handler";
  my $pkg  = $args->{package} || (caller)[0];
  push @IMPORT_FILTERS, [$ext, $desc, $func, $pkg];
}

sub register_io_export {
  my $args = shift || die "Missing arguments";
  my $ext  = lc($args->{extension}) || die "Missing extension";
  my $desc = $args->{description}   || die "Missing description";
  my $func = $args->{handler}       || die "Missing handler";
  my $pkg  = $args->{package} || (caller)[0];
  push @EXPORT_FILTERS, [$ext, $desc, $func, $pkg];
}

sub add_namespace {
  my %used = map { $_ => 1 } @NAMESPACE;
  foreach my $name (@_) {
    push @NAMESPACE, $name if ! $used{$name};
  }
}

sub import_filters {
  &load_filters();
  my @array;
  foreach (sort {$a->[0] cmp $b->[0]} @IMPORT_FILTERS) {
    push @array, [$_->[0], $_->[1]];
  }
  return \@array;
}

sub export_filters {
  &load_filters();
  my @array;
  foreach (sort {$a->[0] cmp $b->[0]} @EXPORT_FILTERS) {
    push @array, [$_->[0], $_->[1]];
  }
  return \@array;
}

### each module loaded will describe its own capabilities
sub load_filters {
  return 1 if $loaded;
  $loaded = 1;

  foreach my $namespace (@NAMESPACE) {
    my $subdir = $namespace;
    $subdir =~ s|::|/|g;
    foreach my $dir (@INC) {
      my $DH;
      next if ! -d "$dir/$subdir";
      if (! opendir($DH,"$dir/$subdir")) {
        warn "Couldn't open \"$dir/$subdir\": $!";
        next;
      }
      while (defined(my $file = readdir($DH))) {
        next if $file !~ /^\w+.pm$/;
        my $filename_short = "$subdir/$file";
        my $filename_long  = "$dir/$filename_short";
        next if $INC{$filename_short}; # require only once
        next if $INC{$filename_long};
        if (! eval {require $filename_short}) {
          warn "Couldn't require $filename_short: $@";
        }
      }
      closedir($DH);
    }
  }

  return 1;
}

###----------------------------------------------------------------###
### public methods

sub new {
  my $class = shift || __PACKAGE__;
  my $self  = shift || [];
  return bless $self, $class;
}

sub run_import {
  my $self = shift;
  my $filter_index = shift;
  my $fh   = shift;
  my $size = shift;

  ### if a filename was passed instead - find the first matching filter_index
  ### this is overload for convenience in writing other modules
  ### PerlCAD will normally only pass the numeric filter index
  if ($filter_index !~ /^\d+$/ && $filter_index =~ /\.(\w+)$/) {
    if (! -e $filter_index || ! -s _) {
      die "Missing file $filter_index (or no size)";
    }
    if (! $fh) {
      $size ||= -s _;
      open($fh, "<", $filter_index) || die "Couldn't open $filter_index for reading";
    }
    my $filters = &import_filters();
    foreach (0 .. $#$filters) {
      next if lc($filters->[$_]->[0]) ne lc($1);
      $filter_index = $_;
      last;
    }
    if ($filter_index !~ /^\d+$/) {
      die "Couldn't find matching filter_index for $filter_index";
    }
  }

  ### now get the method
  my $row = $IMPORT_FILTERS[$filter_index] || die "Invalid filter index $filter_index";
  my $handler = $row->[2] || die "Missing handler";

  ### temporarily become part of the other package, run the import, and return
  my $old_ref = ref $self;
  bless $self, $row->[3];
  $self->$handler($fh, $size);
  bless $self, $old_ref;
}

sub register_hooks {
  my $self = shift;
  my $hash = shift;
  my $map  = &hook_map();
  my $name;
  my $func;
  while (($name, $func) = each %$hash) {
    die "Unknown key $name" if ! defined $map->{$name};
    die "Value must be a subroutine ref for key $name" if ref($func) ne 'CODE';
    $self->[ $map->{$name} ] = $func;
  }
}

sub clear_hooks {
  my $self = shift;
  delete $self->[$_] foreach values %{ &hook_map() };
}

###----------------------------------------------------------------###
### private methods used during import
### call registered methods with the same args passed to us

sub debug {
  my $self = shift;
  return if ! exists $self->[_debug];
  &{ $self->[_debug] };
}

sub data_progress {
  my $self = shift;
  return if ! exists $self->[_data_progress];
  &{ $self->[_data_progress] };
}

sub add_comment {
  my $self = shift;
  return if ! exists $self->[_add_comment];
  &{ $self->[_add_comment] };
}

sub add_variable {
  my $self = shift;
  return if ! exists $self->[_add_variable];
  &{ $self->[_add_variable] };
}

sub add_block {
  my $self = shift;
  return if ! exists $self->[_add_block];
  &{ $self->[_add_block] };
}

sub add_ltype {
  my $self = shift;
  return if ! exists $self->[_add_ltype];
  &{ $self->[_add_ltype] };
}

sub add_layer {
  my $self = shift;
  return if ! exists $self->[_add_layer];
  &{ $self->[_add_layer] };
}

sub add_line {
  my $self = shift;
  return if ! exists $self->[_add_line];
  &{ $self->[_add_line] };
}

sub add_arc {
  my $self = shift;
  return if ! exists $self->[_add_arc];
  &{ $self->[_add_arc] };
}

###----------------------------------------------------------------###

### debug sub - gets run from command line
### useful with perl -d:Profile __FILE__ ~/file/to/import.dxf
### may eventually grow into a more full featured command line operation
if ($0 eq __FILE__ && ! $debug_required) {
  $debug_required = 1;
  my $file = shift(@ARGV) || die "debug mode requires filename as single argument";
  my $io = __PACKAGE__->new;

  ### get the hooks ready
  my $hash = {};
  foreach my $name (keys %{ &hook_map() }) {
    $hash->{$name} = sub { return 0; print "\U$name (".join(",", @_).")\n" };
  }
  $io->register_hooks($hash);

  ### do the import
  $io->run_import($file);
  exit;
}
###----------------------------------------------------------------###

1;
