package PerlCAD::IO::DXF;

###----------------------------------------------------------------###
#  Copyright 2003 - Paul Seamons                                     #
#  Distributed under the GNU General Public License without warranty #
###----------------------------------------------------------------###

use strict;
use PerlCAD::IO;

use Data::DumpEx;

@PerlCAD::IO::DXF::ISA = qw(PerlCAD::IO);
&PerlCAD::IO::register_io_import({
  extension   => 'dxf',
  description => 'Drawing Exchange Format',
  handler     => \&dxf_import_file,
});

sub dxf_import_file {
  my $self = shift;
  my $fh   = shift; # open file handle
  my $size = shift; # size of file

  ### prepare a few subs to pick off lines
  my $read = 0;
  my $N    = 0;

  my $die = sub {
    my $msg = shift;
    my $_N  = shift || $N;
    dex_warn sprintf("$msg at line %d", $_N);
    die sprintf("$msg at line %d", $_N);
  };

  my $get_line = sub {
    my $line1 = <$fh>;
    &$die("Missing data", $N + 1) if ! defined $line1;
    my $line2 = <$fh>;
    &$die("Missing data", $N + 2) if ! defined $line2;
    $N += 2;
    $read += length($line2) + length($line2);
    $line2 =~ s/[\r\n]+$//;
    if ($line1 !~ /^\s*(\d+)\s*$/) {
      &$die("Error parsing dxf - bad number found - $line1 -", $N - 1);
    }
#    warn "F: $1 - $line2\n";
    return ($1, $line2);
  };

  my $get_details = sub {
    my $n;
    my $line;
    my %hash = ();
    while (1) {
      ($n, $line) = &$get_line();
      last if ! $n; # until defined or = 0
      if (exists $hash{$n}) {
#        &$die("Duplicate property found ($n, $line)") if $n != 100;
        $hash{$n} = [$hash{$n}] if ! ref $hash{$n};
        push @{ $hash{$n} }, $line;
      } else {
        $hash{$n} = $line;
      }
    }
    return (\%hash, $n, $line);
  };



  ### now iterate through the lines
  my ($n, $line) = &$get_line();
  my $last = 0;
  my $in_section = '';
  my $in_table   = '';
  my $in_block   = '';
  my $hash;
  $self->debug("# Beginning parse");
  while (defined $n) {
    ### comments
    if ($n == 999) {
      $self->add_comment($line);
      $self->debug("# found a comment ($line)");

    ### begin/end section/table
    } elsif ($n == 0 ) {
      $line = uc($line);
      if ($line eq 'EOF') {
        last;

      ### handle sections
      } elsif ($line eq 'SECTION') {
        &$die("Nested SECTION") if $in_section;
        ($n, $line) = &$get_line();
        &$die("Missing SECTION name") if $n != 2;
        $in_section = uc($line);
        $self->debug("# begin SECTION $in_section");
      } elsif ($line eq 'ENDSEC') {
        &$die("Not in SECTION") if ! $in_section;
        $self->debug("#   end SECTION $in_section");
        $in_section = '';

      ### handle tables
      } elsif ($line eq 'TABLE') {
        &$die("Table outside TABLES SECTION") if $in_section ne 'TABLES';
        &$die("Nested TABLE") if $in_table;
        ($n, $line) = &$get_line();
        &$die("Missing TABLE name ($n, $line)") if $n != 2 || ! $line;
        $in_table = uc($line);
        $self->debug("# begin TABLE $in_table");
      } elsif ($line eq 'ENDTAB') {
        &$die("Not in TABLE") if ! $in_table;
        $self->debug("#   end TABLE $in_table");
        $in_table = '';

      ### handle blocks
      } elsif ($line eq 'BLOCK') {
        &$die("Block outside BLOCKS SECTION") if $in_section ne 'BLOCKS';
        &$die("Nested BLOCK") if $in_block;
        ($hash, $n, $line) = &$get_details();
        my ($name, $desc, $layer, $x, $y, $z) = @$hash{5, 2, 8, 10, 20, 30};
        &$die("Missing BLOCK id (@{[%$hash]})") if ! $name || ! $name;
        $self->add_block('block', $name, $desc, $layer, $x, $y, $z, dex_text($hash));
        $in_block = uc($hash->{5});
        $self->debug("# begin BLOCK id $in_block");
        next; # must do next because of get_details
      } elsif ($line eq 'ENDBLK') {
        &$die("Not in BLOCK") if ! $in_block;
        $self->debug("#   end BLOCK id $in_block");
        $in_block = '';

      ### handle other types
      } else {
        my $type = $line;
        ($hash, $n, $line) = &$get_details();

        if ($type eq 'LTYPE') {
          &$die("Not in LTYPE TABLE ($in_table)") if $in_table ne 'LTYPE';
          my ($name, $desc, $align, $n_items, $length, $dash_length) = @$hash{2, 3, 72, 73, 40, 49};
          $dash_length = join(",", @$dash_length) if ref $dash_length;
          $self->add_ltype($name, $desc, $align, $n_items, $length, $dash_length);

        } elsif ($type eq 'LAYER') {
          &$die("Not in LAYER TABLE ($in_table)") if $in_table ne 'LAYER';
          my ($name, $ltype, $color, $width, $flags) = @$hash{2, 6, 62, 39, 70};
          my $hidden = ($color > 0) ? 1 : 0;
          my $locked = 0;
          $self->add_layer($name, $ltype, abs($color), $width,
                         $hidden, $locked, $flags, dex_text($hash));

        } elsif ($type eq 'LINE' || $type eq '3DLINE') {
          my $blk_id = $in_block || undef;
          my ($ltype, $layer, $color, $width, $x1, $y1, $z1, $x2, $y2, $z2)
            = @$hash{6, 8, 62, 39, 10, 20, 30, 11, 21, 31};
          $self->add_line($blk_id, $ltype, $layer, $color, $width, [$x1, $y1, $z1], [$x2, $y2, $z2]);

        } elsif ($type eq 'CIRCLE') {
          my $blk_id = $in_block || undef;
          my ($ltype, $layer, $color, $width, $x1, $y1, $z1, $radius)
            = @$hash{6, 8, 62, 39, 10, 20, 30, 40};
          $self->add_arc($blk_id, $ltype, $layer, $color, $width, [$x1, $y1, $z1], $radius, 0, 360);

        } elsif ($type eq 'ARC') {
          my $blk_id = $in_block || undef;
          my ($ltype, $layer, $color, $width, $x1, $y1, $z1, $radius, $ang1, $ang2)
            = @$hash{6, 8, 62, 39, 10, 20, 30, 40, 50, 51};
          $self->add_arc($blk_id, $ltype, $layer, $color, $width, [$x1, $y1, $z1], $radius, $ang1, $ang2);

        } else {
          $self->debug("# found a $line");
        }
        next; # must do next because of get_details
      }

    ### import variables
    } elsif ($n == 9) { # variables
      &$die("Not sure of variable format ($line)") if $line !~ /^\$(\w+)$/;
      my $key = $1;
      my($n2,$val) = &$get_line();
      $self->add_variable($key, $val, $in_section);

    } elsif ($n == 62) { # color
    } elsif ($n == 8 ) { # layer
    } elsif ($n == 6 ) { # linetype
    }

    ### get the new line for the next while iteration
    ($n, $line) = &$get_line();

  ### on each row - update our status
  } continue {
    if ($last != int($read / $size * 100)) {
      $last = int($read / $size * 100);
      $self->data_progress($read, $size);
      #if ($dia->wasCancelled) {
      #  dex "Was cancelled";
      #  die "Not implemented";
      #}
    }
  }

#  $dia->setProgress($size);
#  $dia->setTotalSteps(0); # now show the busy cursor


#  $dia->setLabelText("Committing DB...");

  $self->debug("# Ending parse");

  dex $N;
}

1;
