package PerlCAD::Item::Ellipse;

###----------------------------------------------------------------###
#  Copyright 2003 - Paul Seamons                                     #
#  Distributed under the GNU General Public License without warranty #
###----------------------------------------------------------------###

use strict;
use Qt;
use Qt::isa qw(Qt::CanvasPolygonalItem);
use Qt::attributes qw(
                      radius
                      angle1
                      angle2
                      points
                      );
use Data::DumpEx;
use Math::Trig ();

sub NEW {
  my $class = shift;
  my $canvas = shift;
  $class->SUPER::NEW($canvas);

  ### allow initialize of all
  local this->{no_update} = 1;
  setX(shift);
  setY(shift);
  setRadius(shift);
  setAngles(shift, shift);
  update_points();

#
# An ellipse with centre (a,b),
# and axis in x-direction of length c,
# and axis in y-direction of length d,
# has parametric equations x=a+c.cos(t) , y=b + d.sin(t)

# more
#  a = center to side
#  b = center to top
#  e = sqrt(1 - ( b ** 2 / a ** 2 ))
#  c = a * ecen
#  r = a (1 - ecen ** 2) / ( ecen * cos(pheta) )
#  x = c + r * cos(pheta)
#  y = r sing(pheta)
  
}

sub DESTROY {
  hide();
}

sub setX {
  SUPER->setX(@_);
  update_points() if ! this->{no_update};
}

sub setY {
  SUPER->setY(@_);
  update_points() if ! this->{no_update};
}

sub setRadius {
  radius = abs(shift || 0);
  update_points() if ! this->{no_update};
}

sub setAngles {
  my $ang1 = shift;
  my $ang2 = shift;
  ($ang1, $ang2) = ($ang2, $ang1) if $ang1 > $ang2; # ang1 should be less
  die "Invalid angle on angle1 ($ang1)" if $ang1 < 0;
  die "Invalid angle on angle2 ($ang2)" if $ang2 > 360;
  die "Angles may not match ($ang1, $ang2)" if $ang1 == $ang2;
  angle1 = &Math::Trig::deg2rad($ang1);
  angle2 = ($ang2 == 360) ? &Math::Trig::pi2() : &Math::Trig::deg2rad($ang2);
  dex angle1, angle2;
}

sub startPoint {
  return Qt::Point(this->x, this->y);
}

sub update_points {
  invalidate();

  ### optimize the number of points per circle
  my $r = radius;
  my $n = ($r < 5) ? 8 : ($r < 12) ? 16 : ($r < 25) ? 32 : ($r < 75) ? 48 : ($r < 500) ? 64 : 128;
  my @x;
  my @y;
  for (0..$n) {
    my $a = &Math::Trig::pi2() * ($_ / $n) + angle1;
    $a = angle2 if $a > angle2;
    push @x, this->x + $r * cos($a);
    push @y, this->y - $r * sin($a);
    last if $a == angle2;
  }
  points = Qt::PointArray($#x + 1);
  points->setPoint($_, $x[$_], $y[$_]) for 0 .. $#x;

  ### I would use the makeEllipse or makeArc - but they use way too many points
  ### plus they don't close completely
  ### plus QCanvasEllipse doesn't paint the outline
  # points->makeEllipse(this->x - radius, this->y - radius, radius * 2, radius * 2);
  # points->makeArc(this->x - radius, this->y - radius, radius * 2, radius * 2, 0, 360 * 16);
  update();
}

sub rtti {
  1002;
}

sub drawShape {
  my $painter = shift;
  $painter->drawPolyline(points);
#  $painter->setClipRegion(Qt::Region(points));
#  $painter->fillRect(points->boundingRect, brush());
#  $painter->setClipping(0);
}

sub areaPoints {
  return points;
}

sub boundingRect {
  return points->boundingRect;
}

1;
