package PerlCAD::MainWindow;

###----------------------------------------------------------------###
#  Copyright 2003 - Paul Seamons                                     #
#  Distributed under the GNU General Public License without warranty #
###----------------------------------------------------------------###

use Data::DumpEx;

use vars qw($called_from_slot);
use strict;
use Qt;
use Qt::debug;
BEGIN {
  sub set_qt_debug {
    if (shift) { import   Qt::debug qw(ambiguous verbose) }
    else       { unimport Qt::debug }
  }
  &set_qt_debug(1);
}
use Qt::isa qw(Qt::MainWindow);
use Qt::attributes qw(
                      toolbars
                      toolbars_menu
                      scripts_menu
                      doc
                      central_widget
                      central_layout
                      central_splitter
                      config
                      command_box
                      command_win
                      command
                      timer
                      stdin_select
                      dev_no
                      coord_win
                      coord_wgt
                      last_coord
                      labels
                      sig_fig
                      unit
                      prop_layer
                      prop_color
                      prop_ltype
                      prop_width
                      );
use Qt::slots
  slot_modified_toggled  => ['int'],
  slot_filename_changed  => ['QString'],
  slot_sb_toggle_grid    => [],
  slot_sb_toggle_ortho   => [],
  slot_sb_toggle_snap    => [],
  slot_new_coord         => ['double', 'double', 'double'],
  slot_relative_point    => ['double', 'double', 'double'],
  slot_left_screen       => [],
  slot_coord_orientation => ['Qt::Orientation'],
  slot_timeout           => [],
  slot_display_line      => ['QString', 'int', 'int'],
  slot_place_command_win     => ['QDockWindow::Place'],
  ;
use Qt::signals
  sig_process_line => ['QString', 'int'],
  ;
use PerlCAD::Units;
use PerlCAD::View;
use PerlCAD::Doc;
use PerlCAD::Command;
use PerlCAD::CommandBox;
use PerlCAD::Config;
use PerlCAD::UI::LayerCombo;
use PerlCAD::UI::ColorCombo;

BEGIN {
  if ($ENV{REQUIRE_ALL}) {
    require IO::Select;
  }
}

###----------------------------------------------------------------###

sub NEW {
  shift->SUPER::NEW(@_);

#  setUsesBigPixmaps(1);
#  setUsesTextLabel(1);

  init_config();
  init_timer();

  init_commands();
  init_doc();
 
  init_menubar();
  init_statusbar();
  init_toolbars();
  init_stdin();
  init_layout();

  init_views();

  ### little bit larger
  my $resize = Qt::Size(640, 480);
  $resize = $resize->expandedTo(minimumSizeHint());
  resize($resize);
}

sub closeEvent {
  my $event = shift;
  if (doc && doc->is_modified) {
    dex "Need confirm close";
    return 0;
  }

  save_settings();

  SUPER->closeEvent($event);
}

sub save_settings {
#  my $file = Qt::File('/tmp/layout');
#  if ($file->open(2)) {
#    dex;
#    my $stream = Qt::TextStream($file);
#    $stream << this;
#    $file->close();
#  } else {
#    dex;
#  }
}

###----------------------------------------------------------------###

sub init_config {
  return if config;
  config = PerlCAD::Config(this, "configuration");
}

sub init_timer {
  return if timer;
  timer = Qt::Timer(this,'timer');
  timer->start(100); # msec
}

sub init_commands {
  return if command;
  command = PerlCAD::Command(this, "command processor");
  command->load_commands_and_actions;
  command->load_aliases;
}

sub init_menubar {
  my $m = menuBar();
  command->get_action('viewmenubar')->setOn(1);

  ### get the menubar ready
  my $menu = [['&File',
               'filenew',
               'fileopen',
               'filesave',
               'filesaveas',
               '____',
               'fileprint',
               '____',
               'exit'],
              ['&Edit'],
              ['&View',
               'viewgrid',
               'viewortho',
               'viewsnap',
               '____',
               'reload',
               '____',
               ['&Zoom',
                'viewmag',
                'viewmagfit',
                'viewmagx',
                'viewmagxp',
                'viewmag+',
                'viewmag-',
                'viewmagprevious'
                ],
               '____',
               ['&Viewports',
                'view_vports1',
                'view_vports2',
                'view_vports3',
                'view_vports4',
                ],
               ['&Views',
                'viewmenubar',
                'viewstatusbar',
                'viewcommandwin',
                ],
               ],
              ['&Format',
               'colorize',
               'formatlayer',
               ],
              ['&Draw',
               'draw_line',
               'draw_rect',
               ['&Circle',
                'draw_circle_cr',
                'draw_circle_2p',
                'draw_circle_3p',
                ],
               ],
              ['&Modify',
               'eraser',
               ],
              ['&Tools',
               'toolsscript',
               'toolsloadscripts',
               '____',
               ],
              ['&Help',
               'contexthelp',
               ],
              ];

  ### create the menus in order
  my $map = build_bar('menubar', $m, $menu);
  toolbars_menu = $map->{5}; # hate to have these hardcoded
  scripts_menu  = $map->{9};
}

sub init_statusbar {
  my $s = statusBar();

  $s->setSizeGripEnabled(0);
  
  my $a = command->get_action('viewgrid');
  my $w = $s->{bg} = Qt::PushButton("GRID", $s);
  $w->setToggleButton(1);
  $w->setOn($a->isOn);
  this->connect($w, SIGNAL('released()'), SLOT('slot_sb_toggle_grid()'));
  $s->addWidget($w, 0, 1);
  &Qt::WhatsThis::add($w,$a->whatsThis);
  &Qt::ToolTip::add($w,$a->toolTip);

  $a = command->get_action('viewortho');
  $w = $s->{bo} = Qt::PushButton("ORTHO", $s);
  $w->setToggleButton(1);
  $w->setOn($a->isOn);
  this->connect($w, SIGNAL('released()'), SLOT('slot_sb_toggle_ortho()'));
  $s->addWidget($w, 0, 1);
  &Qt::WhatsThis::add($w,$a->whatsThis);
  &Qt::ToolTip::add($w,$a->toolTip);

  $a = command->get_action('viewsnap');
  $w = $s->{bs} = Qt::PushButton("SNAP", $s);
  $w->setToggleButton(1);
  $w->setOn($a->isOn);
  this->connect($w, SIGNAL('released()'), SLOT('slot_sb_toggle_snap()'));
  $s->addWidget($w, 0, 1);
  &Qt::WhatsThis::add($w,$a->whatsThis);
  &Qt::ToolTip::add($w,$a->toolTip);

  command->get_action('viewstatusbar')->setOn(1);
}

sub init_toolbars {

  ### get the toolbar ready
  my $bars = [['File/Edit',
               qw(filenew
                  fileopen
                  filesave
                  fileprint
                  ____
                  undo
                  redo
                  editcut
                  editcopy
                  editpaste
                  find
                  ____
                  contexthelp
                  )],
              ['View',
               qw(viewmag
                  viewmagfit
                  viewmag+
                  viewmag-
                  ____
                  viewgrid
                  viewortho
                  viewsnap
                  )],
              ['Draw',
               qw(draw_line
                  draw_rect
                  draw_circle_cr
                  draw_circle_2p
                  draw_circle_3p
                  )],
              ['Properties'],
              ['Modify',
               qw(eraser
                  )],
              ];

  ### create the toolbars in order
  my $map = build_bar('toolbar', this, $bars);

  ### put them in the list sorted by name
  if (scalar keys %$map) {
    toolbars_menu->insertSeparator;
    my $ref = toolbars = [undef];
    foreach my $n (sort {$map->{$a}->label cmp $map->{$b}->label} keys %$map) {
      my $i = $#$ref + 1;
      $ref->[$i] = $map->{$n};
      my $name = "viewtoolbar_$i";
      command->register_action({
        name      => $name,
        menutext  => "Toolbar" ." &$i - ". $map->{$n}->label,
        is_toggle => 1,
        toggled   => 1,
        slot      => $name,
        method    => sub {
          my $toggle = shift;
          my $on = ($toggle) ? "ON" : "OFF";
          this->slot_process_line("TOOLBAR $i $on");
        },
      });
      command->get_action($name)->addTo(toolbars_menu);
    }
  }

  my $prop_toolbar = $map->{3};
  build_prop_toolbar($prop_toolbar);
}

### auto gen a menubar or toolbar
sub build_bar {
  my $type   = shift;
  my $parent = shift;
  my $array  = shift;
  my $map    = shift || {};
  my $n_ref  = shift || do { my $n = 0; \$n };
  
  foreach my $list (@$array) {
    my $widget = ($type eq 'menubar') ? Qt::PopupMenu(this) : Qt::ToolBar(this);
    $map->{$$n_ref++} = $widget;
    $parent->insertItem($list->[0], $widget) if $type eq 'menubar';
    $widget->setLabel($list->[0]) if $type eq 'toolbar';
    foreach my $i (1 .. $#$list) {
      if (ref($list->[$i])) {
        build_bar($type, $widget, [$list->[$i]], $map, $n_ref);
      } elsif ($list->[$i] =~ /^[-_]+$/) {
        $widget->insertSeparator if $type eq 'menubar';
        $widget->addSeparator    if $type eq 'toolbar';
      } else {
        command->get_action($list->[$i])->addTo($widget);
      }
    }
  }

  return $map;
}

sub build_prop_toolbar {
  my $st = shift;

  ### add the style toolbar items
  my $height = 26;
  my $stw = Qt::HBox($st, 'style toolbar central widget');
  $st->setLabel("Properties");
  $st->setStretchableWidget($stw);
  $st->setHorizontallyStretchable(0);

  my $box;
  my $n;

  prop_layer = PerlCAD::UI::LayerCombo($stw, 'layer combo button', this);
  prop_layer->setMaximumHeight($height);
  my $action = command->get_action('formatlayer');
  if ($action) {
    Qt::ToolTip(prop_layer)->add(prop_layer, $action->toolTip) if $action->toolTip;
    Qt::WhatsThis(prop_layer)->add(prop_layer, $action->whatsThis) if $action->whatsThis;
  }
  command->connect(prop_layer, SIGNAL('clicked()'), SLOT('slot_formatlayer()'));

  prop_color = PerlCAD::UI::ColorCombo($stw, 'color combo', this);
  prop_color->setMaximumHeight($height);
  $action = command->get_action('colorize');
  if ($action) {
    Qt::ToolTip(prop_color)->add(prop_color, $action->toolTip) if $action->toolTip;
    Qt::WhatsThis(prop_color)->add(prop_color, $action->whatsThis) if $action->whatsThis;
  }
  command->connect(prop_color, SIGNAL('clicked()'), SLOT('slot_colorize()'));

  $box = prop_width = Qt::ComboBox($stw);
  $box->setMaximumHeight($height);
  Qt::ToolTip($box)->add($box,"Line Width");
  Qt::WhatsThis($box)->add($box,"Select the width of draw lines");
  $n = 0;
  $box->insertItem("ByLayer", $n ++);
  foreach my $wid (0 .. 8) {
    my $pix = Qt::Pixmap(15,11);
    my $paint = Qt::Painter($pix);
    $paint->setPen(Qt::Pen(Qt::black(), $wid));
    $pix->fill(Qt::white());
    $paint->drawLine(0,6,15,6);
    $box->{ids}->{$n} = $wid;
    $box->insertItem($pix, $wid, $n++);
  }
  $box->setCurrentItem(1);

  $box = prop_ltype = Qt::ComboBox($stw);
  $box->setMaximumHeight($height);
  Qt::ToolTip($box)->add($box,"Line Style");
  Qt::WhatsThis($box)->add($box,"Select the style of draw lines");
  $n = 0;
  $box->insertItem("ByLayer", $n ++);
  foreach my $name (qw(CONTINUOUS HIDDEN CENTER PHANTOM DOT)) {
    my $val = $PerlCAD::DB::LTYPE_MAP{$name};
    my $pix = Qt::Pixmap(48,11);
    my $paint = Qt::Painter($pix);
    $paint->setPen(Qt::Pen(Qt::black(), 0, $val));
    $pix->fill(Qt::white());
    $paint->drawLine(0,6,48,6);
    $box->{ids}->{$n} = $name;
    $box->insertItem($pix, $n ++);
  }
  $box->setCurrentItem(1);
}

sub get_line_info {
  my $i = prop_ltype->currentItem;
  $i = ($i > 0) ? prop_ltype->{ids}->{$i} : 'ByLayer';  

  my $j = prop_color->colorname;

  my $k = prop_width->currentItem;
  $k = ($k > 0) ? prop_width->{ids}->{$k} : 'ByLayer';  

  return [$i, $j, $k];
}

sub get_line_width {
  my $i = prop_width->currentItem;
  my $val = ($i > 0) ? prop_width->{ids}->{$i} : 'ByLayer';
  return [$val, $i];
}

sub set_line_width {
  my $width = shift || 0;
  my $count = prop_width->count;
  return if $count - 1 < $width;
  prop_width->setCurrentItem($width);
}

###----------------------------------------------------------------###

sub toggle_toolbar {
  my $i     = shift;
  my $state = shift;
  my $obj = toolbars->[$i] || return;
  ($state) ? $obj->show : $obj->hide;
  command->get_action("viewtoolbar_$i")->setOn($state);
}

sub init_stdin {
  if (grep {/--command/} @ARGV) {
    @ARGV = grep {! /--command/} @ARGV; # remove it
    this->connect(timer, SIGNAL('timeout()'), SLOT('slot_timeout()'));
    require IO::Select;
    stdin_select = IO::Select->new(\*STDIN);
    command->register_device(this, 'sig_process_line', 'slot_display_line');
  }
}

sub init_doc {
  doc = PerlCAD::Doc(this, "the Document");

  this->connect(doc, SIGNAL('change_filename(QString)'), SLOT('slot_filename_changed(QString)'));
  this->connect(doc, SIGNAL('toggle_modified(int)'    ), SLOT('slot_modified_toggled(int)'));

}

sub get_document {
  return doc;
}

sub get_command {
  return command;
}

### generally this method should not be called until
### a view and layout has been generated
sub load_first_doc {
  if (@ARGV && $ARGV[$#ARGV] && -e $ARGV[$#ARGV]) {
    my $file = pop(@ARGV);
    doc->open_document($file);
  } else {
    doc->new_document;
  }

  command->clear_history();
  command->save_requires_history('POINT', '0,0,0');

#  view->zoom(0,0,12,9);   # set the initial size

  ### allow for some more command line work
  if (stdin_select && @ARGV) {
    emit sig_process_line(shift(@ARGV), dev_no) while @ARGV;
  }
}

###----------------------------------------------------------------###

sub init_views {
  ### someday we could check the config to see if they want more than one at startup
  set_vports(1);
}

sub create_vport {
  my $parent = shift;
  my $name   = shift;
  my $view = PerlCAD::View($parent, $name);

  ### add up ownership
  doc->add_view($view);
  $view->add_doc(doc);
  command->register_device($view, 'sig_process_line', undef, 1);

  ### add in the slots
  $view->connect(timer, SIGNAL('timeout()'), SLOT('slot_timeout()'));
  $view->connect(command, SIGNAL('sig_relative_point(double, double, double)'),
                  SLOT('slot_relative_point(double, double, double)'));
  this->connect($view, SIGNAL('sig_new_coord(double, double, double)'), SLOT('slot_new_coord(double, double, double)'));
  this->connect($view, SIGNAL('sig_left_screen()'), SLOT('slot_left_screen()'));
  command_box->connect($view, SIGNAL('sig_right_mouse()'), SLOT('slot_press_return()'));

  $view->show();
  return $view;
}

sub set_vports {
  my $layout = shift;
  $layout =~ /^(\d)/ || die "Unknown vport layout";
  my $n = $1;
  ### someday we will have 2hor, 2ver, 3right, 3left, 3top, 3bot, 3hor, 3ver

  my $views  = doc->get_views;
  my $n_old  = $#$views + 1;
  die "Invalid number of views ($n)" if $n < 1 || $n > 4;

  return if $n_old == $n;

  ### remove widgets from layout
  if ($n_old == 1) {
    central_layout->remove($views->[0]);
  } elsif ($n_old == 2) {
    central_layout->remove(central_splitter);
    central_splitter->reparent(undef, Qt::Point());
  } elsif ($n_old >= 3) {
    my $s = central_splitter;
    central_layout->remove($s->[0]);
    $s->[$_]->reparent(undef, Qt::Point()) for reverse $#$s .. 0;
  }

  ### delete the views (someday maybe just reuse them)
  ### this does destroy the active view - but - oh well for now
  for my $i (reverse(0 .. $n_old - 1)) {
    $views->[$i]->reparent(undef, Qt::Point());
    doc->remove_view($views->[$i]);
  }

  ### setup the new views
  if ($n == 1) {
    my $view = create_vport(central_widget, "Central View");
    central_layout->insertWidget(0, $view);

  } elsif ($n == 2) {
    central_splitter = Qt::Splitter(&Qt::Horizontal, central_widget, "splitter");
    central_splitter->show();
    central_layout->insertWidget(0, central_splitter);
    create_vport(central_splitter, "Left");
    create_vport(central_splitter, "Right");

  } elsif ($n == 3) {
    my $s = central_splitter = [];
    $s->[0] = Qt::Splitter(&Qt::Horizontal, central_widget, "splitter 0");
    $s->[1] = Qt::Splitter(&Qt::Vertical,   $s->[0], "splitter 1");
    $_->show() foreach @$s;
    central_layout->insertWidget(0, $s->[0]);
    create_vport($s->[1], "LeftTop");
    create_vport($s->[1], "LeftBot");
    create_vport($s->[0], "Right");

  } elsif ($n == 4) {
    my $s = central_splitter = [];
    $s->[0] = Qt::Splitter(&Qt::Horizontal, central_widget, "splitter 0");
    $s->[1] = Qt::Splitter(&Qt::Vertical,   $s->[0], "splitter 1");
    $s->[2] = Qt::Splitter(&Qt::Vertical,   $s->[0], "splitter 2");
    $_->show() foreach @$s;
    central_layout->insertWidget(0, $s->[0]);
    create_vport($s->[1], "LeftTop");
    create_vport($s->[1], "LeftBot");
    create_vport($s->[2], "RightTop");
    create_vport($s->[2], "RightBot");

  } else {
    die "Sorry - No other types available yet";

  }

  ### make one active
  $views->[0]->set_is_active_view(1);
}

###----------------------------------------------------------------###

sub slot_place_command_win {
  if (shift == Qt::DockWindow::InDock()) {
    command_win->setCaption('');
  } else {
    command_win->setCaption("Command");
  }
}

sub init_layout {
  ### set the view holder (can be reset to enable multiple views
  central_widget = Qt::Widget(this);
  setCentralWidget(central_widget);
  central_layout = Qt::HBoxLayout(central_widget, 0, 3, "central_layout");    

  ### setup the command dockwindow
  command_win = Qt::DockWindow(this, "command window");
  my $w = Qt::Widget(command_win, "command central widget");
  command_box = PerlCAD::CommandBox($w,"command_box");
  my $l = Qt::HBoxLayout($w, 0, -1, "command_hbox");
  $l->addWidget(command_box);
  command_win->setWidget($w);
  command_win->setCloseMode(Qt::DockWindow::Undocked());
  this->connect(command_win, SIGNAL('placeChanged(QDockWindow::Place)'),
                SLOT('slot_place_command_win(QDockWindow::Place)'));

  command_win->setVerticallyStretchable(1);
  command_win->setHorizontallyStretchable(1); # expand full line
  command_win->setResizeEnabled(1);           # allow for expansion
  command_win->setNewLine(1);                 # on its own
  if (this->dockWindowsMovable) {
    Qt::MainWindow::moveDockWindow(command_win, Qt::DockBottom());
  }
  command->register_device(command_box, 'sig_process_line', 'slot_display_line', 0);
  command->connect(command_box, SIGNAL('sig_escape_pressed()'), SLOT('slot_escape_pressed()'));
  command_box->connect(command, SIGNAL('sig_send_prompt()'), SLOT('slot_send_prompt()'));
  command_box->set_command(command);
  this->connect(command, SIGNAL('sig_relative_point(double, double, double)'),
                SLOT('slot_relative_point(double, double, double)'));


  ### setup the display of coordinates
  my $ref = labels = {};
  sig_fig = 3;
  unit = 'NONE';
  coord_win = Qt::ToolBar(this, "coordinates window");
  coord_wgt = Qt::Widget(coord_win, "Holder widget");
  coord_win->setStretchableWidget(coord_wgt);
  coord_win->setHorizontallyStretchable(1); # expand full line
  coord_win->setNewLine(1);                 # on its own
  $ref->{rel_t}   = Qt::Label("Rel:",   coord_wgt);
  $ref->{rel_l}   = Qt::Label("",       coord_wgt, "relative label");
  $ref->{coord_t} = Qt::Label("Coord:", coord_wgt);
  $ref->{coord_l} = Qt::Label("",       coord_wgt, "coord label x,y");
  slot_coord_orientation(Qt::Horizontal());

}

sub slot_new_coord {
  last_coord = \@_;
  last_coord ||= 0;
  update_coord();
}

sub slot_relative_point {
  my ($x, $y, $z) = @_;
  labels->{relative_point} = [$x, $y, $z];
  update_coord();
}

sub update_coord {
  my ($x, $y, $z) = @{ last_coord() || [0,0,0] };
  labels->{coord_l}->setText(&PerlCAD::Units::ucs_to_unit($x, unit, sig_fig)
                             . ", "
                             . &PerlCAD::Units::ucs_to_unit($y, unit, sig_fig));

  use Math::Trig qw(rad2deg);
  labels->{relative_point} ||= [0, 0, 0];
  my $X =  $x - labels->{relative_point}->[0];
  my $Y =  $y - labels->{relative_point}->[1];
  my $Z =  $z - labels->{relative_point}->[2];
  my $sig = sig_fig;
  my $len = sqrt($X ** 2 + $Y ** 2 + $Z ** 2);
  my $deg = rad2deg(atan2($Y, $X));
  labels->{rel_l}->setText(&PerlCAD::Units::ucs_to_unit($len, unit, sig_fig)
                           . " < "
                           . &PerlCAD::Units::ucs_to_unit(($deg < 0) ? 360 + $deg : $deg, 'DEGREE', sig_fig));
}


sub slot_left_screen {
  labels->{coord_l}->setText("");
  labels->{rel_l}->setText("");
}

sub slot_coord_orientation {
  my $o = shift;
  my $l = ($o eq Qt::Horizontal())
    ? Qt::HBoxLayout(coord_wgt, 0, -1, "coord_hbox")
    : Qt::VBoxLayout(coord_wgt, 0, -1, "coord_vbox");
  $l->addWidget(labels->{$_}, /_l/ ? 1 : 0) for qw(coord_t coord_l rel_t rel_l);
}

###----------------------------------------------------------------###

sub set_dev_no {
  dev_no = shift;
}

sub slot_timeout {
  if (stdin_select) {
    while (my ($fh) = stdin_select->can_read(0)) {
      my $line = scalar <$fh>;
      emit sig_process_line($line, dev_no);
    }
  }
}

sub slot_display_line {
  my $message = shift;
  return if ! stdin_select;
  return if dev_no == shift;
  print STDOUT "$message\n";
}

###----------------------------------------------------------------###

sub slot_modified_toggled {
  _set_caption();
}

sub slot_filename_changed {
  _set_caption();
}

sub _set_caption {
  my $caption = shift;
  if (! $caption) {
    $caption = doc->get_filename;
    $caption .= " (modified)" if doc->is_modified;
    $caption .= " - PerlCAD";
  }
  setCaption($caption);
}

###----------------------------------------------------------------###

sub toggle_menubar {
  my $state = shift;
  command->get_action('viewmenubar')->setOn($state);
  $state ? menuBar()->show : menuBar()->hide;
}

sub toggle_statusbar {
  my $state = shift;
  command->get_action('viewstatusbar')->setOn($state);
  $state ? statusBar()->show : statusBar()->hide;
}

sub toggle_commandwin {
  my $state = shift;
  command->get_action('viewcommandwin')->setOn($state);
  $state ? command_win->show : command_win->hide;
}

sub toggle_grid {
  my $state = shift;
  statusBar()->{bg}->setOn($state);
  command->get_action('viewgrid')->setOn($state);
  doc->active_view->set_grid($state);
}

sub toggle_ortho {
  my $state = shift;
  statusBar()->{bo}->setOn($state);
  command->get_action('viewortho')->setOn($state);
  doc->active_view->set_ortho($state);
}

sub toggle_snap {
  my $state = shift;
  statusBar()->{bs}->setOn($state);
  command->get_action('viewsnap')->setOn($state);
  doc->active_view->set_snap($state);
}

sub slot_sb_toggle_grid {
  ### doing this will call the action which will call the command
  ### which will call toggle_grid
  my $a = command->get_action('viewgrid');
  $a->setOn($a->isOn ? "" : 1);
}

sub slot_sb_toggle_ortho {
  my $a = command->get_action('viewortho');
  $a->setOn($a->isOn ? "" : 1);
}

sub slot_sb_toggle_snap {
  my $a = command->get_action('viewsnap');
  $a->setOn($a->isOn ? "" : 1);
}

###----------------------------------------------------------------###

1;

__END__
#    begin                : Thu Feb 21 20:29:30 MST 2002
#    copyright            : (C) 2002 by Paul T Seamons
  setCaption(doc->URL().fileName(),false);

void KdraftApp::saveOptions()
{	
  config->setGroup("General Options");
  config->writeEntry("Geometry", size());

  config->writeEntry("Show Main Toolbar", viewFileBar->isChecked());
  config->writeEntry("MainBarPos", (int) toolBar("mainToolBar")->barPos());

  config->writeEntry("CoordBarPos", (int) toolBar("coordBar")->barPos());

  config->writeEntry("Show Command Toolbar", viewCommandBar->isChecked());
  config->writeEntry("CommandBarPos", (int) toolBar("commandBar")->barPos());

  config->writeEntry("Show Statusbar",viewStatusBar->isChecked());

  fileOpenRecent->saveEntries(config,"Recent Files");
}
void KdraftApp::readOptions()
{
	
  config->setGroup("General Options");

  bool bView;
  bView = config->readBoolEntry("Show Statusbar", true);
  viewStatusBar->setChecked(bView);
  slotViewStatusBar();
	

  bView = config->readBoolEntry("Show Main Toolbar", true);
  viewFileBar->setChecked( bView );
  slotViewFileBar();

  bView = config->readBoolEntry("Show Command Toolbar", true);
  viewCommandBar->setChecked( bView );
  slotViewCommandBar();


  // bar position settings
  KToolBar::BarPosition toolBarPos;
  toolBarPos=(KToolBar::BarPosition) config->readNumEntry("MainBarPos", KToolBar::Top);
  toolBar("mainToolBar")->setBarPos(toolBarPos);

  toolBarPos=(KToolBar::BarPosition) config->readNumEntry("CoordBarPos", KToolBar::Top);
  toolBar("coordBar")->setBarPos(toolBarPos);

  toolBarPos=(KToolBar::BarPosition) config->readNumEntry("CommandBarPos", KToolBar::Top);
  toolBar("commandBar")->setBarPos(toolBarPos);

  // initialize the recent file list
  fileOpenRecent->loadEntries(config,"Recent Files");

  QSize size=config->readSizeEntry("Geometry");
  if(!size.isEmpty())
  {
    resize(size);
  }
}

void KdraftApp::saveProperties(KConfig *_cfg)
{
  if(doc->URL().fileName()!=i18n("Untitled") && !doc->isModified())
  {
    // saving to tempfile not necessary

  }
  else
  {
    KURL url=doc->URL();	
    _cfg->writeEntry("filename", url.url());
    _cfg->writeEntry("modified", doc->isModified());
    QString tempname = kapp->tempSaveName(url.url());
    QString tempurl= KURL::encode_string(tempname);
    KURL _url(tempurl);
    doc->saveDocument(_url);
  }
}


void KdraftApp::readProperties(KConfig* _cfg)
{
  QString filename = _cfg->readEntry("filename", "");
  KURL url(filename);
  bool modified = _cfg->readBoolEntry("modified", false);
  if(modified)
  {
    bool canRecover;
    QString tempname = kapp->checkRecoverFile(filename, canRecover);
    KURL _url(tempname);
  	
    if(canRecover)
    {
      doc->openDocument(_url);
      doc->setModified();
      setCaption(_url.fileName(),true);
      QFile::remove(tempname);
    }
  }
  else
  {
    if(!filename.isEmpty())
    {
      doc->openDocument(url);
      setCaption(url.fileName(),false);
    }
  }
}		

