package PerlCAD::View;

###----------------------------------------------------------------###
#  Copyright 2003 - Paul Seamons                                     #
#  Distributed under the GNU General Public License without warranty #
###----------------------------------------------------------------###

use vars qw(%COLORMAP);
use strict;
use Qt;
use Qt::isa qw(Qt::CanvasView);
use Qt::attributes qw(doc
		      dev_no
                      canvas
                      bg_color
                      grid
                      grid_on
                      grid_space_x
                      grid_space_y
                      ortho_on
                      snap_on
                      snap_space_x
                      snap_space_y
                      x1 y1 z1 x2 y2 z2
                      previous_wcs1
                      previous_wcs2
                      dx_per_pixel
                      last_point
                      crosshairs
                      lines
                      arcs
                      reload_queue
                      relative_point
                      mover_objects
                      mover_color
                      selected_objects
                      selected_color
                      draw_sample_on
                      draw_sample_objs
                      draw_sample_sub
                      draw_sample_state
                      coord
                      last_reload
                      );
#use Qt::GlobalSpace qw(bitBlt);
use Qt::slots
  slot_toggle_ortho   => [],
  slot_toggle_snap    => [],
  slot_toggle_grid    => [],
  slot_timeout        => [],
  slot_relative_point => ['double', 'double', 'double'],
  slot_new_data       => ['int', 'bool', 'bool'],
  ;
use Qt::signals
  sig_ortho_toggled => ['int'],
  sig_snap_toggled  => ['int'],
  sig_grid_toggled  => ['int'],
  sig_new_coord     => ['double', 'double', 'double'],
  sig_left_screen   => [],
  sig_right_mouse   => [],
  sig_mid_mouse     => [],
  sig_new_screen_dimensions => ['double', 'double', 'double', 'double'],
  sig_process_line => ['QString', 'int'],
  ;

use constant ch_on    => 1;     # crosshairs on
use constant ch_cross => 4;     # width of crosshair cross
use constant ch_start => 10;    # the rest of the crosshair
use constant ch_z     => 10_000_000; # z index of crosshair
use constant rp_dia   => 7;     # relative point diamond radius
use constant sel_box  => 2;     # radius of the selection box
use constant show_sel_box => 0; # show the selection box
use constant zoom_border => .05; # extra space on edges of zoom

### these two numbers control the reload process
### they affect the user's perception of reload
use constant RELOAD_N_BETWEEN_UPDATE => 20;
use constant RELOAD_N_BETWEEN_QUEUE  => 200;

### standard canvas items
use PerlCAD::Item::Line;
use PerlCAD::Item::Arc;
use PerlCAD::Item::CrossHair;
use PerlCAD::Color qw(aci_to_name);
use PerlCAD::Coord ();
use PerlCAD::Dump qw(debug);

use Data::DumpEx;

###----------------------------------------------------------------###

sub NEW {
  my $class  = shift;
  my $parent = shift;
  my $name   = shift;
  $class->SUPER::NEW($parent, $name);

  setMinimumWidth(100);
  setMinimumHeight(50);

  ### turn on mouse tracking for the canvas - both are necessary
  ### for a QCanvasView
  setMouseTracking(1);
  viewport()->setMouseTracking(1);

  ### we will handle moving around the canvas
  ### turn off the scrollbars
  setVScrollBarMode(Qt::ScrollView::AlwaysOff());
  setHScrollBarMode(Qt::ScrollView::AlwaysOff());

  ### this will hold the objects
  canvas = Qt::Canvas(10, 10);
  setCanvas(canvas);

  ### allow for multiple viewports to show the active
  setFrameShape(&Qt::Frame::Panel);
  setFrameShadow(&Qt::Frame::Sunken);
  setLineWidth(1);

  ### color settings
  set_black_on_white(0);
  mover_color    = Qt::Color(Qt::yellow());
  selected_color = Qt::Color(Qt::red());

  ### crosshairs (cursor) and relative_point
  crosshairs = [];
  my $pen = Qt::Pen(Qt::Color(Qt::darkGray()));
  my $n   = (! ch_on) ? 0 : show_sel_box ? 9 : 5;
  for my $i (0 .. $n) {
    crosshairs->[$i] = PerlCAD::Item::CrossHair(canvas);
    crosshairs->[$i]->setPen($pen);
    crosshairs->[$i]->setZ(ch_z);
    crosshairs->[$i]->show;
  }
  relative_point = {};
  my $ref = relative_point->{items} = [];
  $pen = Qt::Pen(Qt::Color(Qt::yellow()));
  for my $i (0 .. 4) {
    $ref->[$i] = Qt::CanvasLine(canvas);
    $ref->[$i]->setPen($pen);
    $ref->[$i]->setZ(ch_z - 1);
    $ref->[$i]->show;
  }


  ### init other variables
  set_snap_space(1, 1);
  set_grid_space(1, 1);

  zoom(0, 0, 12, 9, 1);
  set_vpoint([0, 0, 1]);
}

sub add_doc {
  doc = shift;
}

sub set_dev_no {
  dev_no = shift;
}

sub get_dev_no {
  return dev_no;
}

sub send_process_line {
  my $line = shift;
  emit sig_process_line($line, dev_no);
}

sub set_is_active_view {
  my $on = shift;
  doc->set_active_view(this) if $on;
  if ($on) {
    if (doc->count_views() == 1) {
      setFrameShape(&Qt::Frame::NoFrame);
    } else {
      setFrameShadow(&Qt::Frame::Raised);
    }
  } else {
    setFrameShadow(&Qt::Frame::Sunken);
  }
  update();
}

###----------------------------------------------------------------###

sub set_black_on_white {
  my $on = shift;
  bg_color     = $on ? Qt::white() : Qt::black();
  $COLORMAP{7} = $on ? Qt::black() : Qt::white();

  canvas->setBackgroundColor(bg_color);

  if (ch_on) {
    setCursor(Qt::Cursor(Qt::BlankCursor()));
  }
}

sub is_black_on_white {
  return bg_color == Qt::black();
}

sub set_grid {
  grid_on = shift;
  update_grid();
}

sub set_snap {
  snap_on = shift;
}

sub set_ortho {
  ortho_on = shift;
}

sub set_snap_space {
  snap_space_x = shift;
  snap_space_y = shift || snap_space_x;
}

sub set_grid_space {
  grid_space_x = shift;
  grid_space_y = shift || snap_space_y;
}

###----------------------------------------------------------------###

sub wcs_to_point {
#  my ($x, $y, $z) = @{ coord->project(shift) };
  my ($x, $y, $z) = @{ shift() };
  my $p = Qt::Point;
  $p->setX( int(($x - x1) * width() / (x2 - x1)) );
#  $p->setY( height() - int($y * height() / (y2 - y1)) );
  $p->setY( height() - int(($y - y1) * height() / (y2 - y1)) );
#  dex "W ($y - ".y2.",".y1.") => (".$p->y.")";
  return $p;
}

sub point_to_wcs {
  my $point = shift;
  my $x = $point->x / width() * (x2 - x1) + x1;
  my $y = (height() - $point->y) * (y2 - y1) / height() + y1;
  my $z = 0;
  return [$x, $y, $z];
}

sub snap_wcs {
  my ($x, $y, $z) = @{ shift() };
  my $recurse = shift;
  if (draw_sample_on && ortho_on) {
#    my ($start_x, $start_y) = @{ point_to_wcs(last_point) };
    my ($start_x, $start_y) = @{ relative_point->{xyz} };
    if (abs($start_x - $x) > abs($start_y - $y)) {
      $y = $start_y;
      if (snap_on) {
        my $_x = snap_space_x * int($x / snap_space_x);
        $x = ($x - $_x > snap_space_x / 2) ? $_x + snap_space_x : $_x;
      }
    } else {
      $x = $start_x;
      if (snap_on) {
        my $_y = snap_space_y * int($y / snap_space_y);
        $y = ($y - $_y > snap_space_y / 2) ? $_y + snap_space_y : $_y;
      }
    }
  } elsif (snap_on) {
    my $_x = snap_space_x * int($x / snap_space_x);
    $x = ($x - $_x > snap_space_x / 2) ? $_x + snap_space_x : $_x;
    my $_y = snap_space_y * int($y / snap_space_y);
    $y = ($y - $_y > snap_space_y / 2) ? $_y + snap_space_y : $_y;
  }

  return [$x,$y,$z];
}

###----------------------------------------------------------------###

sub set_vpoint {
  my $point = shift;
  coord ||= PerlCAD::Coord->new;
  coord->set_vpoint($point);
}

sub zoom {
  my $x1 = shift;
  my $y1 = shift;
  my $x2 = shift;
  my $y2 = shift;
  my $no_reload = shift;
  my $no_zoom_border = shift;
  $_ ||= 0 foreach $x1, $x2, $y1, $y2;

  ### lowest to highest
  ($x1, $x2) = ($x2, $x1) if $x1 > $x2;
  ($y1, $y2) = ($y2, $y1) if $y1 > $y2;

  ### fit map to coordinates
  my $ratio_ucs = ($y2 - $y1) / ($x2 - $x1);
  my $ratio_scr = height() / width();
  if ($ratio_ucs < $ratio_scr) {
    $y2 = $y1 + abs($x2 - $x1) * $ratio_scr;
  } else {
    $x2 = $x1 + abs($y2 - $y1) / $ratio_scr;
  }

  if (! $no_zoom_border) {
    ### allow for zoom border
    my $zbx = ($x2 - $x1) * zoom_border;
    my $zby = ($y2 - $y1) * zoom_border;
    if ($zbx > $zby) {
      $x1 -= $zby;
      $y1 -= $zby;
      $x2 += 2 * $zbx - $zby;
      $y2 += $zby;
    } else {
      $x1 -= $zbx;
      $y1 -= $zbx;
      $x2 += $zbx;
      $y2 += 2 * $zby - $zbx;
    }
  }

  ### store results
  previous_wcs1 = [x1, y1, z1];
  previous_wcs2 = [x2, y2, z2];
  x1 = $x1;
  y1 = $y1;
  x2 = $x2;
  y2 = $y2;
  ### store the ratio so that window resizes can honor the old scale
  dx_per_pixel = ($x2 - $x1) / width();

  emit sig_new_screen_dimensions($x1, $y1, $x2, $y2);
  reload_queue = undef;
  reload() if ! $no_reload;
  update_grid();
  update_relative_point();
  update_selection();
}

sub zoom_extents {
  my ($x1, $y1, $z1, $x2, $y2, $z2) = @{ doc->get_extents };
  zoom($x1, $y1, $x2, $y2);
}

sub zoom_previous {
  zoom(previous_wcs1->[0], previous_wcs1->[1],
       previous_wcs2->[0], previous_wcs2->[1],
       );
}

sub zoom_by_x {
  my $zx = shift || 2;
  my $cenx = (x1 + x2) / 2;
  my $ceny = (y1 + y2) / 2;
  my $dx   = (x2 - x1) / $zx / 2;
  my $dy   = (y2 - y1) / $zx / 2;
  zoom($cenx - $dx, $ceny - $dy, $cenx + $dx, $ceny + $dy);
}

sub zoom_by_xp {
  my $z = shift || 50;
  zoom_by_x($z / 100);
}

sub zoom_in {
  zoom_by_x(2);
}

sub zoom_out {
  zoom_by_x(.5);
}

###----------------------------------------------------------------###

sub resizeEvent {
  my $event = shift;

  SUPER->resizeEvent($event);

  canvas->resize(width(), height());

  update_crosshairs();

  ### may want to work on the first resize to not query objects
  zoom(x1, y1, dx_per_pixel * width(), dx_per_pixel * height(), 0, 1);
}

###----------------------------------------------------------------###

sub leaveEvent {
  emit sig_left_screen();
  return if ! ch_on;
  crosshairs->[$_]->hide for 0 .. 5;
  canvas->update;
}

sub enterEvent {
  return if ! ch_on;
  crosshairs->[$_]->show for 0 .. 5;
}

sub mousePressEvent {
  my $event = shift;

  &set_is_active_view(1);
  
  if ($event->button == Qt::RightButton()) {
    set_draw_sample(undef);
    emit sig_right_mouse();
    return;
  } elsif ($event->button == Qt::MidButton()) {
    emit sig_mid_mouse();
    return;
  }


  my $point_e = $event->pos;
  my $wcs_s   = snap_wcs(point_to_wcs($point_e));
  my $point_s = wcs_to_point($wcs_s);
  my $sig_fig = 3; #this->parent->sig_fig;
  $_ = sprintf("%.${sig_fig}f",$_) foreach @$wcs_s;
  send_process_line("$wcs_s->[0],$wcs_s->[1],$wcs_s->[2]");
  return;
}

sub mouseReleaseEvent {
  my $event = shift;

  my $wcs = snap_wcs(point_to_wcs($event->pos));
#  emit sigMouseLeftRelease(newC);
}

sub mouseMoveEvent {
  my $event = shift;
  my $wcs = snap_wcs(point_to_wcs($event->pos));
  emit sig_new_coord(@$wcs);

  last_point = wcs_to_point($wcs);

  $event->accept();

  update_crosshairs();

  update_draw_sample() if draw_sample_on;

  update_selection();
}

###----------------------------------------------------------------###

sub slot_toggle_ortho {
  ortho_on = ! ortho_on;
  emit sig_ortho_toggled(ortho_on);
}
sub slot_toggle_snap {
  snap_on = ! snap_on;
  emit sig_snap_toggled(snap_on);
}
sub slot_toggle_grid {
  grid_on = ! grid_on;
  emit sig_grid_toggled(grid_on);
}

###----------------------------------------------------------------###

sub update_grid {
  my $pixmap  = Qt::Pixmap(canvas->width, canvas->height);
  my $do_grid = 1;

  if (grid_on && defined x1) {
    my $dx = grid_space_x;
    my $dy = grid_space_y;

    my $start_x = $dx * int((x1 - $dx) / $dx);
    my $end_x   = $dx * int((x2 + $dx) / $dx);
    if (($end_x - $start_x) / $dx > width() / 4) {
      send_process_line("# Grid spacing too small in the x direction");
      $do_grid = 0;
      return;
    }

    my $start_y = $dy * int((y1 - $dy) / $dy);
    my $end_y   = $dy * int((y2 + $dy) / $dy);
    if (($end_y - $start_y) / $dy > height() / 4) {
      send_process_line("# Grid spacing too small in the y direction");
      $do_grid = 0;
      return;
    }

    if ($do_grid) {
      my $painter = Qt::Painter($pixmap);
      $painter->setPen(Qt::Pen(Qt::darkGray()));
      for (my $x = $start_x; $x < $end_x; $x += $dx) {
        next if $x < x1 || $x > x2;
        for (my $y = $start_y; $y < $end_y; $y += $dy) {
          next if $y < y1 || $y > y2;
          $painter->drawPoint(wcs_to_point([$x, $y, 0]));
        }
      }
    }
  }

  canvas->setBackgroundPixmap($pixmap);
}

###----------------------------------------------------------------###

### make the crosshair thing
sub update_crosshairs {
  return if ! ch_on;

  my $p = last_point || return;
  crosshairs->[0]->setPoints($p->x, $p->y - ch_start, $p->x,   0);
  crosshairs->[1]->setPoints($p->x, $p->y + ch_start, $p->x,   height());
  crosshairs->[2]->setPoints($p->x - ch_start, $p->y, 0,           $p->y);
  crosshairs->[3]->setPoints($p->x + ch_start, $p->y, width(),     $p->y);
  crosshairs->[4]->setPoints($p->x+0, $p->y + ch_cross, $p->x+0, $p->y - ch_cross);
  crosshairs->[5]->setPoints($p->x + ch_cross, $p->y+0, $p->x - ch_cross, $p->y+0);

  if (show_sel_box) {
    crosshairs->[6]->setPoints($p->x - sel_box, $p->y - sel_box, $p->x + sel_box, $p->y - sel_box);
    crosshairs->[7]->setPoints($p->x + sel_box, $p->y + sel_box, $p->x + sel_box, $p->y - sel_box);
    crosshairs->[8]->setPoints($p->x + sel_box, $p->y + sel_box, $p->x - sel_box, $p->y + sel_box);
    crosshairs->[9]->setPoints($p->x - sel_box, $p->y - sel_box, $p->x - sel_box, $p->y + sel_box);
  }

  canvas->update;
}

sub update_selection {
  my $p = last_point || return;
  my $new_mover = {};
  my $old_mover = mover_objects; # mouse over
  my $rect = Qt::Rect($p->x - sel_box, $p->y - sel_box, sel_box * 2 + 1, sel_box * 2 + 1);
  my $list = canvas->collisions($rect);
  foreach my $item (@$list) {
    my $id = $item->{id} || next;
    $item->confirm_collision($p->x, $p->y, sel_box) || next;
    if ($old_mover->{$id}) {
      $new_mover->{$id} = delete($old_mover->{$id});
    } else {
      $new_mover->{$id} = [$item, $item->pen, $item->z];
      my $pen = Qt::Pen($item->pen);
      $pen->setColor(mover_color);
      $item->setPen($pen);
      $item->setZ(ch_z);
      $item->update;
    }
    last; # eventually we can allow for cycling through the other items
  }
  while ($_ = each %$old_mover) {
    my ($item, $pen, $z) = @{ $old_mover->{$_} };
    $item->setPen($pen);
    $item->setZ($z);
    $item->update;
  }
  mover_objects = $new_mover;
}

sub slot_relative_point {
  relative_point->{xyz} = \@_;
  update_relative_point();
}

sub update_relative_point {
  return if ! relative_point->{xyz};
  my $point = wcs_to_point(relative_point->{xyz});
  my $ref = relative_point->{items};
  $ref->[0]->setPoints($point->x - rp_dia/2, $point->y + rp_dia/2, $point->x, $point->y);
  $ref->[1]->setPoints($point->x - rp_dia, $point->y, $point->x, $point->y + rp_dia);
  $ref->[2]->setPoints($point->x + rp_dia, $point->y, $point->x, $point->y + rp_dia);
  $ref->[3]->setPoints($point->x + rp_dia, $point->y, $point->x, $point->y - rp_dia);
  $ref->[4]->setPoints($point->x - rp_dia, $point->y, $point->x, $point->y - rp_dia);
  canvas->update;
}

sub set_draw_sample {
  draw_sample_objs  = [];
  if ($_[0]) {
    draw_sample_sub   = shift;
    draw_sample_state = shift;
    draw_sample_on    = 1;
    update_draw_sample();
  } else {
    draw_sample_on    = undef;
    canvas->update;
  }
}

sub update_draw_sample {
  return if ! draw_sample_on;
  my @queue = @{ draw_sample_state() };
  push @queue, snap_wcs(point_to_wcs(last_point));
               #point_to_wcs(last_point);
  my $code_ref = draw_sample_sub;
  my $objs  = eval { &$code_ref(draw_sample_state, @queue) };
  debug "$@" if $@;
  if (! $objs) {
    draw_sample_objs = [];
    canvas->update;
    return;
  }
  
  my $pen = Qt::Pen(Qt::green());
  for (my $i = 0; $i <= $#$objs; $i ++) {
    my $row  = $objs->[$i];
    my $type = uc($row->[0]);
    my $obj  = draw_sample_objs->[$i];
    if (! $obj) {
      if ($type eq 'LINE') {
        $obj = PerlCAD::Item::Line(canvas);
      } elsif ($type eq 'ARC') {
        $obj = PerlCAD::Item::Arc(canvas);
      } else {
        die;
      }
      draw_sample_objs->[$i] = $obj;
    }
    if ($type eq 'LINE') {
      my $p1 = wcs_to_point($row->[1]);
      my $p2 = wcs_to_point($row->[2]);
      $obj->setPoints($p1->x, $p1->y, $p2->x, $p2->y);
    } elsif ($type eq 'ARC') {
      my $p1 = wcs_to_point($row->[1]);
      my $rad = $row->[2]  / dx_per_pixel;
      #my $rad = wcs_to_point([$row->[2], 0, 0])->x;
      $obj->set_all($p1, $rad, $row->[3], $row->[4]);
    } else {
      die;
    }
    $obj->setPen($pen);
    $obj->setZ(ch_z + 2);
    $obj->show;
  }

  canvas->update;
}

###----------------------------------------------------------------###

### this will requery for objects and redraw them as necessary
sub reload {
  return if ! doc;

  if (! defined last_reload) {
    last_reload = doc->query_last_mod();
  }

  my $end = shift || 0;
  my $start = last_reload || 0;
  if (! $end) { # reload all objects
    $end = last_reload;
    $start = 0;
  } else {
    last_reload = $end;
  }

  my $h = height();
  my $w = width();

  ### The reload_queue is intended to buffer the updates drawn to the screen.
  ### If a resize occurs - it will kindly timeout after n objects to redraw the screen.
  ### Without the reload_queue all objects get drawn before resize occurs.
  ### If we used qt's db cursors there may be a way around this.
  if (! reload_queue) {
    if (! $start) {
      lines = [];
      arcs  = [];
    }
    reload_queue = {
      lines => sub { doc->query_lines($start, $end, x1,y1,x2,y2) },
      arcs  => sub { doc->query_arcs($start, $end, x1,y1,x2,y2) },
    };
  }

  ### draw lines
  if (reload_queue->{lines}) {
    if (! reload_queue->{lines_cursor}) {
      my $get_cursor = reload_queue->{lines};
      reload_queue->{lines_cursor} = &$get_cursor();
    }
    my $cursor = reload_queue->{lines_cursor};
    $cursor ||= sub { return undef };
    my $n = 0;
    while (defined (my $ref = &$cursor())) {
      draw_line(@$ref);
      canvas->update if ! (++ $n % RELOAD_N_BETWEEN_UPDATE); # affects reload speed
      return if $n >= RELOAD_N_BETWEEN_QUEUE; # affects resizing
    }
    canvas->update;
    delete reload_queue->{lines};
  }


  ### draw arcs
  if (reload_queue->{arcs}) {
    if (! reload_queue->{arcs_cursor}) {
      my $get_cursor = reload_queue->{arcs};
      reload_queue->{arcs_cursor} = &$get_cursor();
    }
    my $cursor = reload_queue->{arcs_cursor};
    $cursor ||= sub { return undef };
    my $n = 0;
    while (defined (my $ref = &$cursor())) {
      draw_arc(@$ref);
      canvas->update if ! (++ $n % RELOAD_N_BETWEEN_UPDATE); # affects reload speed
      return if $n >= RELOAD_N_BETWEEN_QUEUE; # affects resizing
    }
    canvas->update;
    delete reload_queue->{arcs};
  }


#  warn "Done with reload";
  reload_queue = undef;
}

sub slot_new_data {
  my $last_mod = shift;
  my $has_lines = shift;
  my $has_arcs  = shift;
  reload($last_mod);
}

sub draw_line {
  my ($id, $blk, $ltype, $color, $layer, $width) = @_;
  my $p1 = wcs_to_point([@_[6..8]]);
  my $p2 = wcs_to_point([@_[9..11]]);
  ### should work out way to determine clip plane of view window
  my $line = PerlCAD::Item::Line(canvas, $id);
  $line->setPen(get_pen($ltype, $color, $width));
  $line->setPoints($p1->x, $p1->y, $p2->x, $p2->y);
  $line->setZ($id) if $id;
  $line->show;
  push @{ lines() }, $line;
}

sub draw_arc {
  my ($id, $blk, $ltype, $color, $layer, $width, $x1, $y1, $z1, $rad, $ang1, $ang2) = @_;
  my $p1 = wcs_to_point([$x1, $y1, $z1]);
  $rad = $rad / dx_per_pixel; #my $p2 = wcs_to_point([$rad, 0, 0]);
  my $arc = PerlCAD::Item::Arc(canvas, $id, $p1->x, $p1->y, $rad, $ang1, $ang2);
  $arc->setPen(get_pen($ltype, $color, $width));
  $arc->show;
  push @{ arcs() }, $arc;
}  

### allow for cleanup actions
sub slot_timeout {
  if (reload_queue) {
    reload();
  }
}

sub get_pen {
  my $ltype = shift;
  my $color = uc(shift);
  my $width = shift;
  if (! $COLORMAP{$color}) {
    my $name;
    if ($color eq 'BYLAYER' || $color eq 'BYBLOCK'
        || $color eq '0' || $color eq '256') {
      $name = "white"; # hack - should do a lookup
    } elsif ($color =~ /^\d+$/) {
      $name = aci_to_name($color);
    } else {
      $name = $color;
    }
    $COLORMAP{$color} = Qt::Color();
    $COLORMAP{$color}->setNamedColor($name);
  }
  my $pen = Qt::Pen($COLORMAP{$color});
  $pen->setStyle($ltype);
  $pen->setWidth($width || 0);
  return $pen;
}

###----------------------------------------------------------------###

1;
