#!/usr/bin/perl -w

###----------------------------------------------------------------###
#  Copyright 2003 - Paul Seamons                                     #
#  Distributed under the GNU General Public License without warranty #
###----------------------------------------------------------------###

=head1 NAME

make_dist.pl - For automatically creating tar.gz files.

=head1 DESCRIPTION

This script is roughly similar to running the command "make dist".

It is intented to be used for doing the initial tar.gz (dist) creation.
It generally not intended to be run after the distribution has been
delivered - in those cases it is normally sufficent to type "make dist"
to recreate the tar.gz (using make dist will result in the loss of the
revision number though).

This script was generated to help in some environments where the source files
for a distribution may reside in multiple locations that may vary from
their location in the destination tar.  It also allows for
autogeneration of the MANIFEST, updating of a rpm spec file, updating
of release numbers, removal of development "use lib"'s in executable
files, and basic checking of module compilation.

=head1 ARGUMENTS
 
Arguments are positional.  They are base_dir and tar_dir.  Both
have reasonable defaults.

=over 4

=item C<base_dir>

This is where the base of the repository is found for determing
all other path names.  If no base_dir is given, this script will
assume that it resides in /home/user/Package or in
/home/user/Package/build (if the repository was /home/userPackage).
It such a case it uses __FILE__ to resolve the respository.

=item C<tar_dir>

This represents the directory where the finished tar files


   ~/make_dist.pl ; # shell expands to home
### arguments are base_dir and tar_dir
### base_dir is where this repository runs from
### tar_dir is where the resulting tar.gz will go

=back

=head1 CONFIGURATION

Configuration is done by direct modification of this file.
Most of the items that can be configured are located at the top.
There is also a section at the bottom called EXTRA_SUBS that 
is a good location for running any additional code before writing
the MANIFEST.

=cut


use strict;
use File::Path qw(rmtree);
use File::Find qw(find);
use File::Copy qw(copy);

###----------------------------------------------------------------###
### CONFIGURATION

### tar distribution name
my $NAME = 'perlcad';

### used to get $VERSION and perldoc
my $INFO_FROM = 'PerlCAD::Main';

### any modules in addition to $INFO_FROM to try and require
my @OTHER_VERIFY = qw(
                      PerlCAD::Icon
                      );

### any files to exclude
my %EXCLUDE = (
               'perlcad_splash.png' => 1,
               'perlcad_splash.xcf' => 1,
               );

### the list of files to include in the MANIFEST file
### normally - this will be populated automatically
my %MANIFEST = ();

###----------------------------------------------------------------###

### where is this repository
### try and figure it out from our runtime - it may also be specified
### as the first argument to this program
my $base_dir = shift;
if (! $base_dir) {
  $base_dir = (__FILE__ =~ m|^(.+?)/(build/)?[^/]+$|) ? $1 : die "invalid base";
  $base_dir =~ s{^\.\./*}{
    my $pwd = `pwd`;
    $pwd =~ s|/[^/]+$||;
    "$pwd/"; # return of s///
  }e;
  $base_dir =~ s{^\./*}{
    my $pwd = `pwd`;
    chomp $pwd;
    "$pwd/"; # return of s///
  }e;
  $base_dir =~ s|/+$||;
}
print "Building from current view of $base_dir\n";


### create a temporary build directory
my $build_dir = "$base_dir/tmp_build";
if (-d $build_dir) {
  rmtree($build_dir) || die "Couldn't remove $build_dir for build: $!";
}
mkdir($build_dir, 0755) || die "Couldn't create $build_dir for build: $!";


### where to finished tar.gz files go
my $DIST = shift || "$base_dir/tgz";

### $SRC and $DEST represent directories (groups) for files to copy
my $SRC = {
  lib   => "$base_dir/lib",
  bin   => "$base_dir/bin",
  img   => "$base_dir/img",
  doc   => "$base_dir/doc",
  test  => "$base_dir/t",
  build => "$base_dir/build",
};

my $DEST = {
  lib   => "$build_dir/lib",
  bin   => "$build_dir/bin",
  img   => "$build_dir/img",
  doc   => "$build_dir/doc",
  test  => "$build_dir/t",
  build => "$build_dir",
};

###----------------------------------------------------------------###

### loop on the groups to copy over files
foreach my $group (keys %$SRC) {
  my $src  = $SRC->{$group};
  my $dest = $DEST->{$group};

  if (! -d $src) {
    warn "No such directory $src for $group on build - skipping\n";
    next;
  }
  if (! -d $dest) {
    mkdir($dest, 0755) || die "Couldn't mkdir $dest for $group on build: $!";
  }
  find(sub {
    return 0 if /^\./; # no hidden files or directories
    return 0 if $File::Find::name =~ /CVS/;
    return 0 if /~$/; # no temp files
    return 0 if $EXCLUDE{$_};
    my $short = ($File::Find::name =~ m|\Q$src\E/*(.+)$|) ? $1 : return 0;
    if (-d) {
      mkdir("$dest/$short", 0755) || die "Couldn't mkdir $dest/$short: $!";
    } else {
      copy("$src/$short", "$dest/$short")
        || die "Couldn't copy $src/$short to $dest/$short: $!";
      $MANIFEST{"$dest/$short"} = 1;
    }
  }, $src);
}

### remove any development links from the bin directory
find(sub {
  my $file = $File::Find::name;
  return 0 if ! $MANIFEST{$file};
  my ($atime,$mtime) = (stat $file)[8,9];
  open (IN, "+<$file") || die "Couldn't open $file program: $!";
  my $text  = join "", <IN>;
  my $count = $text =~ s/^use lib.+\n//g;
  if ($count) {
    seek  IN, 0, 0;
    print IN $text;
    truncate IN, length($text);
    close IN;
  }
  utime($atime, $mtime, $file) || die "Couldn't restore time: $!";
}, $DEST->{bin}) if -d $DEST->{bin};


### test compile some sources
unshift @INC, $DEST->{lib};
chdir($DEST->{lib}) || die "Couldn't chdir to $DEST->{lib}: $!";
foreach my $pkg ($INFO_FROM, @OTHER_VERIFY) {
  my $file = $pkg;
  $file .= ".pm" if $file !~ /\.\w+$/;
  $file =~ s|::|/|g;
  if (! eval {require $file}) {
    die "Sources don't compile for $pkg - aborting ($@)\n";
  }
}
shift @INC;


### get the version number
### the version will come from $VERSION in $INFO_FROM
no strict 'refs';
my $version = ${"${INFO_FROM}::VERSION"} || die "No \$VERSION in $INFO_FROM";
use strict 'refs';

### the release is incremented based upon previously build versions
### if there is no directory found we will assume release 1
my $release = 0;
if (-d $DIST) {
  opendir(DH, $DIST) || die "Couldn't open $DIST: $!";
  while (defined(my $sub = readdir(DH))) {
    next if $sub !~ /^\Q$NAME\E-\Q$version\E-(\d+)\.tar\.gz$/;
    $release = $1 if $1 > $release;
  }
  closedir(DH);
}
$release ++;


### get the pod - if any - to be used in README and SPEC
my $file = $INFO_FROM;
$file .= ".pm" if $file !~ /\.\w+$/;
$file =~ s|::|/|g;
my $pod = `pod2text $DEST->{lib}/$file`;
$pod = '' if $pod =~ /^No \s+ Documentation \s+ found/xi;


### write out the README if none found
my $readme = "$DEST->{build}/README";
if (! -e $readme && $pod) {
  open (OUT, ">$readme") || die "Couldn't open $readme: $!";
  print OUT $pod;
  close OUT;
  $MANIFEST{$readme} = 1;
}


### update the spec file if there is one
### we will swap out any @DESCRIPTION@ style
### tags with those sections from the perldoc
my $spec = "$DEST->{build}/$NAME.spec";
if (-e $spec) {
  my %POD = $pod =~ /^([A-Z]+) \s+ (?s: (.+?) ) \s+ (?: $ | (?= ^[A-Z]+))/mxg;
  $POD{SUMMARY} = delete $POD{NAME};
  $POD{VERSION} = $version;
  $POD{RELEASE} = $release;
  my($atime,$mtime) = (stat $spec)[8,9];
  open (MOD, "+<$spec") || die "Couldn't open $spec: $!";
  my $text2 = join "", <MOD>;
  if ($text2 =~ s/\@([A-Z]+)\@/$POD{$1} || ''/eg) {
    seek MOD, 0, 0;
    print MOD $text2;
    truncate MOD, length($text2);
    close MOD;
  }
  utime($atime,$mtime,$spec) || die "Couldn't utime $spec: $!";
}
  

###----------------------------------------------------------------###
### EXTRA_SUBS
### Module specific extras can be run here
### This should be done before the writing of the MANIFEST

### generate static icons
#mkdir("$DEST->{img}/icons",0755) || die "Couldn't mkdir icons: $!";
#my $list = &PerlCAD::Icon::write_icons_to_directory("$DEST->{img}/icons");
#$MANIFEST{$_} = 1 foreach @$list;

###----------------------------------------------------------------###


### write out the MANIFEST
my $manifest = "$DEST->{build}/MANIFEST";
if (! -e $manifest) {
  $MANIFEST{$manifest} = 1;
  open (OUT, ">$manifest") || die "Couldn't open $manifest: $!";
  foreach my $file (sort keys %MANIFEST) {
    $file =~ s|^\Q$build_dir/\E||og;
    print "File $file included in MANIFEST\n";
    print OUT "$file\n";
  }
  close OUT;
}


### rename the temp build direcotry
my $tarname = "${NAME}-${version}-${release}";
rmtree("$base_dir/$tarname") if -e "$base_dir/$tarname";
rename($build_dir, "$base_dir/$tarname") || die "Couldn't move $build_dir: $!";


### tar it up
chdir($base_dir) || die "Couldn't chdir to $base_dir: $!";
`tar -zcvf $DIST/$tarname.tar.gz $tarname`;
print "Wrote out $DIST/$tarname.tar.gz\n";


### all done - remove the sources
rmtree("$base_dir/$tarname");


1;

###----------------------------------------------------------------###
## Build a menu item
#cat > ./usr/lib/menu/%{lname} <<EOF
#?package(%{lname}):\
#command="%{_appsbindir}/%{lname}"\
#title="%{name}"\
#longtitle="%{name} Drafting Program"\
#needs="x11"\
#icon="%{lname}.png" \
#section="Applications/Publishing"
#EOF
#
